# -*- coding: utf-8 -*-
# © Copyright EnterpriseDB UK Limited 2011-2025
#
# This file is part of Barman.
#
# Barman is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Barman is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Barman.  If not, see <http://www.gnu.org/licenses/>

import collections
import filecmp
import logging
import os
import shutil
from abc import ABCMeta, abstractmethod
from distutils.version import LooseVersion as Version
from glob import glob
from tempfile import NamedTemporaryFile

from barman import output, xlog
from barman.cloud_providers import ObjectKeyAlreadyExists
from barman.command_wrappers import CommandFailedException, PgReceiveXlog
from barman.compression import InternalCompressor, compression_registry
from barman.exceptions import (
    AbortedRetryHookScript,
    ArchiverFailure,
    CompressionException,
    DuplicateWalFile,
    MatchingDuplicateWalFile,
)
from barman.hooks import HookScriptRunner, RetryHookScriptRunner
from barman.infofile import WalFileInfo
from barman.remote_status import RemoteStatusMixin
from barman.utils import fsync_dir, fsync_file, mkpath, with_metaclass
from barman.xlog import is_partial_file

_logger = logging.getLogger(__name__)


class WalArchiverQueue(list):
    def __init__(self, items, errors=None, skip=None, batch_size=0, total_size=0):
        """
        A WalArchiverQueue is a list of WalFileInfo which has two extra
        attribute list:

        * errors: containing a list of unrecognized files
        * skip: containing a list of skipped files.

        It also stores batch run size information in case
        it is requested by configuration, in order to limit the
        number of WAL files that are processed in a single
        run of the archive-wal command.

        .. note::
            This class was originally designed to hold all the WAL files pending to be
            archived, and to process up to a certain number of files (the batch size).
            However, it is now used to hold only the files that are being processed in a
            single run of the archive-wal command. That change was made to avoid the
            overhead of having to create lots of WalFileInfo objects, even if only a
            subset of them would be used by the archiver in a given batch. We might want
            to rework or remove this class in the future, as it doesn't seem to add much
            value to the archiving process anymore.

        :param items: iterable from which initialize the list
        :param errors: an optional list of unrecognized files
        :param skip: an optional list of skipped files
        :param batch_size: size of the current batch run (0=unlimited)
        :param total_size: the total number of WAL files available for archiving.
        """
        super(WalArchiverQueue, self).__init__(items)
        self.skip = []
        self.errors = []
        if skip is not None:
            self.skip = skip
        if errors is not None:
            self.errors = errors
        # Normalises batch run size
        if batch_size > 0:
            self.batch_size = batch_size
        else:
            self.batch_size = 0

        self.total_size = total_size
        self.run_size = len(self)


class WalStorageStrategy(metaclass=ABCMeta):
    """
    Abstract base class for WAL storage strategies.

    WAL storage strategies are used to effectively store WAL files in the
    connfigured destination, be it local filesystem or cloud storage.
    """

    def __init__(self, backup_manager, server):
        """
        Constructor.

        :param barman.backup_manager.BackupManager backup_manager: The backup manager
        :param barman.server.Server server: The server
        """
        self.backup_manager = backup_manager
        self.config = backup_manager.config
        self.server = server

    @abstractmethod
    def save(self, compressor, encryption, wal_info, **kwargs):
        """
        Effectively persist a WAL file according to the configured destination.

        :param compressor: the compressor for the file (if any)
        :param None|Encryption encryption: the encryptor for the file (if any)
        :param WalFileInfo wal_info: the WAL file is being processed
        :param kwargs: additional parameters for the storage strategy, if any
        """

    @abstractmethod
    def delete(self, wals_to_delete):
        """
        Delete WAL files according to the configured destination.

        :param dict[str, list[WalFileInfo]] wals_to_delete: A dictionary where key is
            the WAL directory name and value is a list of wal_info objects representing
            the WALs to be deleted in that directory.
        :return list[str]: a list of deleted WAL names.
        """

    def _run_pre_archive_scripts(self, wal_info, src_file):
        """
        Run the pre-archive scripts.

        :param WalFileInfo wal_info: the WAL file info object
        :param str src_file: the source WAL file path
        """
        # Run the pre_archive_script if present
        script = HookScriptRunner(self.backup_manager, "archive_script", "pre")
        script.env_from_wal_info(wal_info, src_file)
        script.run()

        # Run the pre_archive_retry_script if present
        retry_script = RetryHookScriptRunner(
            self.backup_manager, "archive_retry_script", "pre"
        )
        retry_script.env_from_wal_info(wal_info, src_file)
        retry_script.run()

    def _run_post_archive_scripts(self, wal_info, dest_file, error):
        """
        Run the post-archive scripts.

        :param WalFileInfo wal_info: the WAL file info object
        :param str dest_file: the destination WAL file path
        :param Exception|None error: the exception raised during
            the archival process, if any
        """
        try:
            retry_script = RetryHookScriptRunner(
                self.backup_manager, "archive_retry_script", "post"
            )
            retry_script.env_from_wal_info(wal_info, dest_file, error)
            retry_script.run()
        except AbortedRetryHookScript as e:
            # Ignore the ABORT_STOP as it is a post-hook operation
            _logger.warning(
                "Ignoring stop request after receiving "
                "abort (exit code %d) from post-archive "
                "retry hook script: %s",
                e.hook.exit_status,
                e.hook.script,
            )

        # Run the post_archive_script if present
        script = HookScriptRunner(self.backup_manager, "archive_script", "post", error)
        script.env_from_wal_info(wal_info, dest_file)
        script.run()

    def _run_pre_delete_wal_scripts(self, wal_info):
        """
        Run the pre-delete hook-scripts, if any, on the given WAL.

        :param barman.infofile.WalFileInfo wal_info: WAL to run the script on.
        """
        # Run the pre_wal_delete_script if present.
        script = HookScriptRunner(self.backup_manager, "wal_delete_script", "pre")
        script.env_from_wal_info(wal_info)
        script.run()
        # Run the pre_wal_delete_retry_script if present.
        retry_script = RetryHookScriptRunner(
            self.backup_manager, "wal_delete_retry_script", "pre"
        )
        retry_script.env_from_wal_info(wal_info)
        retry_script.run()

    def _run_post_delete_wal_scripts(self, wal_info, error=None):
        """
        Run the post-delete hook-scripts, if any, on the given WAL.

        :param barman.infofile.WalFileInfo wal_info: WAL to run the script on.
        :param None|str error: error message in case a failure happened.
        """
        # Run the post_wal_delete_retry_script if present.
        try:
            retry_script = RetryHookScriptRunner(
                self.backup_manager, "wal_delete_retry_script", "post"
            )
            retry_script.env_from_wal_info(wal_info, None, error)
            retry_script.run()
        except AbortedRetryHookScript as e:
            # Ignore the ABORT_STOP as it is a post-hook operation
            _logger.warning(
                "Ignoring stop request after receiving "
                "abort (exit code %d) from post-wal-delete "
                "retry hook script: %s",
                e.hook.exit_status,
                e.hook.script,
            )
        # Run the post_wal_delete_script if present.
        script = HookScriptRunner(self.backup_manager, "wal_delete_script", "post")
        script.env_from_wal_info(wal_info, None, error)
        script.run()


class LocalWalStorageStrategy(WalStorageStrategy):
    """
    WAL storage strategy for local filesystem storage.

    .. note::
        For now, this class is also responsible for encrypting and compressing files
        before storing them, but in the future this responsibility should be moved
        elsewhere, desirably in a shared component, so that strategies' only
        responsibility is to actually persist files.
    """

    def __init__(self, backup_manager, server):
        super(LocalWalStorageStrategy, self).__init__(backup_manager, server)
        self.files_to_remove = []

    def _check_duplicate(self, src_file, dst_file, wal_info):
        """
        Check if the destination WAL file already exists in local storage, and if so,
        whether it is identical to the source file.

        :param str src_file: the source WAL file path
        :param str dst_file: the destination WAL file path
        :param WalFileInfo wal_info: the WAL file info object
        :raise DuplicateWalFile: if the destination file exists and is
            different from the source file
        :raise MatchingDuplicateWalFile: if the destination file exists and is
            identical to the source file
        """
        if not os.path.exists(dst_file):
            return

        comp_manager = self.backup_manager.compression_manager
        dst_info = self.backup_manager.get_wal_file_info(dst_file)
        src_uncompressed = src_file
        dst_uncompressed = dst_file
        try:
            # If the existing destination file is already encrypted, it can't be
            # decrypted or uncompressed to perform any of the later comparisons
            # (because we cannot assume the encryption passphrase is always
            # available in the configuration).
            if dst_info.encryption:
                raise DuplicateWalFile(wal_info)
            # If the existing file is already compressed, decompress it to a
            # <dst_wal_path>.uncompressed file
            if dst_info.compression is not None:
                dst_uncompressed = dst_file + ".uncompressed"
                comp_manager.get_compressor(dst_info.compression).decompress(
                    dst_file, dst_uncompressed
                )
            # If the source file is already compressed (because the user
            # compressed it manually with a script in the archive_command),
            # then decompress it to a <src_wal_path>.uncompressed file
            if wal_info.compression:
                src_uncompressed = src_file + ".uncompressed"
                comp_manager.get_compressor(wal_info.compression).decompress(
                    src_file, src_uncompressed
                )
            # Directly compare files. When the files are identical raise a
            # MatchingDuplicateWalFile exception, otherwise raise DuplicateWalFile
            if filecmp.cmp(dst_uncompressed, src_uncompressed):
                raise MatchingDuplicateWalFile(wal_info)
            else:
                raise DuplicateWalFile(wal_info)
        finally:
            if src_uncompressed != src_file:
                os.unlink(src_uncompressed)
            if dst_uncompressed != dst_file:
                os.unlink(dst_uncompressed)

    def _compress_file(self, compressor, src_file, dst_dir, wal_info):
        """
        Compress *src_file* to a "temp" file inside *dst_dir* and updates *wal_info*.

        .. note::
            The temporary compressed file will have the same name as the source file,
            with a ".compressed" suffix appended. It's created on the assumption that
            the caller will later remove the file or rename/move it to its final
            destination.

        :param compressor: the compressor to use
        :param str src_file: the file to compress
        :param str dst_dir: the directory where the compressed file will be created
        :param WalFileInfo wal_info: the WAL file info object
        :returns: the path to the compressed temporary file
        """
        tmp_file = "%s.compressed" % os.path.join(dst_dir, os.path.basename(src_file))
        compressor.compress(src_file, tmp_file)
        self.files_to_remove.append(src_file)
        wal_info.compression = compressor.compression
        return tmp_file

    def _encrypt_file(self, encryption, src_file, dst_dir, wal_info):
        """
        Encrypt *src_file* to a "temp" file inside *dst_dir* and updates *wal_info*.

        .. note::
            The temporary encrypted file will have the same name as the source file,
            with a ".gpg" suffix appended. It's created on the assumption that
            the caller will later remove the file or rename/move it to its final
            destination.

        :param encryption: the encryptor to use
        :param str src_file: the file to encrypt
        :param str dst_dir: the directory where the encrypted file will be created
        :param WalFileInfo wal_info: the WAL file info object
        :returns: the path to the encrypted temporary file
        """
        # The encrypted file will have the .gpg extension
        encrypted_file = encryption.encrypt(src_file, dst_dir)
        self.files_to_remove.append(src_file)
        wal_info.encryption = encryption.NAME
        return encrypted_file

    def _copy_stats(self, src_file, dst_file, wal_info):
        """
        Copy stats from *src_file* to *dst_file* and updates its *wal_info*.

        This is used preserve the metadata of the original file after compression or
        encryption, updating only the size in *wal_info* accordingly.

        :param str src_file: the source WAL file path
        :param str dst_file: the destination WAL file path
        :param WalFileInfo wal_info: the WAL file info object
        """
        shutil.copystat(src_file, dst_file)
        stat = os.stat(dst_file)
        wal_info.size = stat.st_size

    def _rename_or_copy_file(self, src_file, dst_file):
        """
        Rename or copy *src_file* to *dst_file*.

        Rename is attempted first, and if it fails (because the source and
        destination are on different filesystems), a copy is performed.

        :param str src_file: the source file path
        :param str dst_file: the destination file path
        """
        try:
            os.rename(src_file, dst_file)
        except OSError:
            shutil.copy2(src_file, dst_file)
            self.files_to_remove.append(src_file)

    def _remove_intermediary_files(self):
        """Remove any intermediary files created during the archival process"""
        for file in self.files_to_remove:
            try:
                os.unlink(file)
            except OSError as e:
                _logger.warning("Could not remove intermediary file %s: %s", file, e)

    def _fsync_contents(self, src_dir, dst_dir, dst_file):
        """
        Fsync the contents of source and destination directories and the
        destination file.

        :param str src_dir: the source directory path
        :param str dst_dir: the destination directory path
        :param str dst_file: the destination file path
        """
        # Fsync the destination file
        fsync_file(dst_file)
        # Fsync the source directory to ensure the removal of the original file
        fsync_dir(src_dir)
        # Fsync the target directory to ensure the presence of the new file
        fsync_dir(dst_dir)

    def save(self, compressor, encryption, wal_info, **kwargs):
        """
        Effectively persist a WAL file according to the configured destination.

        Compression and encryption are applied, if requested.

        :param compressor: the compressor for the file (if any)
        :param None|Encryption encryption: the encryptor for the file (if any)
        :param WalFileInfo wal_info: the WAL file is being processed
        :param kwargs: additional parameters for the storage strategy, if any
        :raises DuplicateWalFile: if the destination file exists and is
            different from the source file
        :raises MatchingDuplicateWalFile: if the destination file exists and is
            identical to the source file
        """
        src_file = wal_info.orig_filename
        src_dir = os.path.dirname(src_file)
        dst_file = wal_info.fullpath(self.server)
        dst_dir = os.path.dirname(dst_file)
        mkpath(dst_dir)

        current_file = src_file
        error = None
        try:
            self._run_pre_archive_scripts(wal_info, current_file)
            self._check_duplicate(current_file, dst_file, wal_info)

            if compressor and not wal_info.compression:
                current_file = self._compress_file(
                    compressor, current_file, dst_dir, wal_info
                )
            if encryption:
                current_file = self._encrypt_file(
                    encryption, current_file, dst_dir, wal_info
                )

            # Update stats, in case compression/encryption changed the current file
            if src_file != current_file:
                self._copy_stats(src_file, current_file, wal_info)

            # Perform the real filesystem operation with the xlogdb lock taken
            # This makes the operation atomic from the xlogdb's perspective
            with self.server.xlogdb("a") as fxlogdb:
                self._rename_or_copy_file(current_file, dst_file)
                self._remove_intermediary_files()
                self._fsync_contents(src_dir, dst_dir, dst_file)
                # At this point the original file has been removed
                wal_info.orig_filename = None
                fxlogdb.write(wal_info.to_xlogdb_line())

        except Exception as e:
            error = e
            raise

        finally:
            self._run_post_archive_scripts(wal_info, dst_file, error)

    def delete(self, wals_to_delete):
        wals_deleted = []
        for wal_dir, wal_list in wals_to_delete.items():
            delete_directory = False
            # Each directory can contain up to 256 WAL files. If the deletion list
            # contains 256 entries, the entire directory can be safely deleted
            # Otherwise, check if all WALs in the directory are in the deletion list
            if len(wal_list) >= 256:
                delete_directory = True
            else:
                wal_names_to_delete = {wal_info.name for wal_info in wal_list}
                wal_names_in_dir = os.listdir(wal_dir)
                if set(wal_names_in_dir).issubset(wal_names_to_delete):
                    delete_directory = True
            # If the directory can be deleted, run the hook-scripts on each WAL file
            # before and after the rmtree. Otherwise, delete each WAL individually
            if delete_directory:
                self._delete_wal_directory(wal_dir, wal_list)
                wals_deleted.extend(wal_info.name for wal_info in wal_list)
            else:
                for wal_info in wal_list:
                    self._delete_wal_file(wal_info)
                    wals_deleted.append(wal_info.name)

        return wals_deleted

    def _delete_wal_directory(self, wal_dir, wal_list):
        for wal_info in wal_list:
            self._run_pre_delete_wal_scripts(wal_info)
        shutil.rmtree(wal_dir)
        for wal_info in wal_list:
            self._run_post_delete_wal_scripts(wal_info)

    def _delete_wal_file(self, wal_info):
        """
        Perform the actual deletion of the WAL file from local storage.

        :param WalFileInfo wal_info: the WAL file info object
        """
        self._run_pre_delete_wal_scripts(wal_info)
        error = None
        try:
            os.unlink(wal_info.fullpath(self.server))
            try:
                os.removedirs(os.path.dirname(wal_info.fullpath(self.server)))
            except OSError:
                # This is not an error condition
                # We always try to remove the trailing directories,
                # this means that hashdir is not empty.
                pass
        except OSError as e:
            error = "Ignoring deletion of WAL file %s for server %s: %s" % (
                wal_info.name,
                self.config.name,
                e,
            )
            output.warning(error)

        self._run_post_delete_wal_scripts(wal_info, error)

    def exists(self, wal_full_path):
        return os.path.exists(wal_full_path)

    def get_full_path(self, wal_name):
        # Build the path which contains the file
        hash_dir = os.path.join(self.config.wals_directory, xlog.hash_dir(wal_name))
        # Build the WAL file full path
        full_path = os.path.join(hash_dir, wal_name)
        return full_path


class CloudWalStorageStrategy(WalStorageStrategy):
    """
    WAL storage strategy for cloud storage.
    """

    def __init__(self, backup_manager, server):
        super(CloudWalStorageStrategy, self).__init__(backup_manager, server)
        self.cloud_interface = self.server.get_wal_cloud_interface()

    def _get_compression_extension(self, compression):
        """
        Return the file extension for the given compression algorithm.

        The *compression* value must be either ``None`` or a key present in
        :data:`~barman.compression.compression_registry` whose class defines an
        ``EXTENSION`` attribute. Passing an unrecognised algorithm or one without
        ``EXTENSION`` (e.g. ``pigz``, ``custom``) will raise ``KeyError`` or
        ``AttributeError`` respectively.

        :param str|None compression: the compression algorithm name
        :return: the file extension (e.g. ".gz") or "" if no compression
        :rtype: str
        """
        if not compression:
            return ""
        return compression_registry[compression].EXTENSION

    def _build_wal_object_key(self, wal_name, compression):
        """
        Build the full cloud object key for a WAL file.

        :param str wal_name: the WAL file name (e.g. ``000000010000000000000001``)
        :param str|None compression: the compression algorithm name used for the WAL
        :return: the full cloud object key, including compression extension if applicable
        :rtype: str
        """
        ext = self._get_compression_extension(compression)
        return (
            os.path.join(
                self.cloud_interface.path,
                self.config.name,
                "wals",
                xlog.hash_dir(wal_name),
                wal_name,
            )
            + ext
        )

    def _check_duplicate(self, wal_info, object_key):
        """
        Check if the cloud object *object_key* is identical to the source file.

        If the WAL was compressed, the ``decompress`` parameter of ``download_file``
        is used to decompress the cloud object before comparison.

        :param WalFileInfo wal_info: the WAL file info object
        :param str object_key: the cloud storage object key
        :raise DuplicateWalFile: if the destination file exists and is
            different from the source file
        :raise MatchingDuplicateWalFile: if the destination file exists and is
            identical to the source file
        """
        with NamedTemporaryFile(delete=True) as downloaded_file:
            self.cloud_interface.download_file(
                object_key, downloaded_file.name, decompress=wal_info.compression
            )
            if filecmp.cmp(wal_info.orig_filename, downloaded_file.name):
                raise MatchingDuplicateWalFile(wal_info)
            else:
                raise DuplicateWalFile(wal_info)

    def save(self, compressor, encryption, wal_info, **kwargs):
        """
        Effectively persist a WAL file according to the configured destination.

        If a *compressor* is provided, the WAL file is compressed in-memory before
        upload and the cloud key includes the compression extension. The *wal_info*
        object is updated with the compression type and compressed size so that
        xlogdb records accurate metadata.

        :param compressor: an :class:`~barman.compression.InternalCompressor` instance,
            or ``None`` if no compression is desired
        :param None|Encryption encryption: the encryptor for the file (if any)
        :param WalFileInfo wal_info: the WAL file is being processed
        :param kwargs: additional parameters for the storage strategy, if any:

            * "skip_delete": if ``True``, the source file will not be deleted after a
              successful upload.

        :raises CompressionException: if *compressor* is not an
            :class:`~barman.compression.InternalCompressor` instance

        .. note::
            Only ``InternalCompressor`` subclasses (gzip, bzip2, xz, zstd, lz4, snappy)
            are supported for cloud WAL storage. ``pigz`` and ``custom`` are not
            supported because they rely on external processes and cannot compress
            in-memory.

        .. note::
            Encryption is not yet supported for cloud WAL storage. The *encryption*
            parameter is kept for interface compatibility.
        """
        if compressor is not None and not isinstance(compressor, InternalCompressor):
            raise CompressionException(
                "Cloud WAL storage only supports in-memory compression algorithms "
                "(gzip, bzip2, xz, zstd, lz4, snappy). "
                "pigz and custom compressors are not supported."
            )
        error = None
        try:
            self._run_pre_archive_scripts(wal_info, wal_info.orig_filename)
            # Compress if a compressor is provided.
            upload_fileobj = None
            if compressor:
                # The source file is read and closed before uploading so we don't hold
                # and/or leak the fd.
                with open(wal_info.orig_filename, "rb") as fileobj:
                    upload_fileobj = compressor.compress_in_mem(fileobj)
                wal_info.compression = compressor.compression
                wal_info.size = upload_fileobj.getbuffer().nbytes
            else:
                upload_fileobj = open(wal_info.orig_filename, "rb")

            with upload_fileobj:
                with self.server.xlogdb("a") as fxlogdb:
                    key = self._build_wal_object_key(
                        wal_info.name, wal_info.compression
                    )
                    try:
                        self.cloud_interface.upload_fileobj(
                            fileobj=upload_fileobj, key=key, fail_if_exists=True
                        )
                    except ObjectKeyAlreadyExists:
                        self._check_duplicate(wal_info, key)
                    fxlogdb.write(wal_info.to_xlogdb_line())
        except Exception as e:
            error = e
            raise
        finally:
            self._run_post_archive_scripts(wal_info, wal_info.orig_filename, error)

        if not kwargs.get("skip_delete", False):
            os.unlink(wal_info.orig_filename)
            wal_info.orig_filename = None

    def delete(self, wals_to_delete):
        wal_objects_to_delete = []
        for _, wal_list in wals_to_delete.items():
            for wal_info in wal_list:
                object_key = self._build_wal_object_key(
                    wal_info.name, wal_info.compression
                )
                wal_objects_to_delete.append(object_key)
                self._run_pre_delete_wal_scripts(wal_info)

        self.cloud_interface.delete_objects(wal_objects_to_delete)

        wals_deleted = []
        for _, wal_list in wals_to_delete.items():
            for wal_info in wal_list:
                self._run_post_delete_wal_scripts(wal_info)
                wals_deleted.append(wal_info.name)

        return wals_deleted

    def exists(self, wal_full_path):
        return self.cloud_interface.check_object_existence(wal_full_path)

    def get_full_path(self, wal_name):
        """
        Construct the full cloud object key for a given WAL file name.

        .. note::
            This method uses the current compression configuration to determine
            the file extension. If the compression config has changed since the WAL
            was originally stored, this method will return a path that does not match
            the actual cloud object. Callers that need to handle WALs stored under a
            previous compression config should use bucket listing instead.

        :param str wal_name: the WAL file name
        :return: the full cloud object key
        :rtype: str
        """
        return self._build_wal_object_key(wal_name, self.config.compression)


class WalArchiver(with_metaclass(ABCMeta, RemoteStatusMixin)):
    """
    Base class for WAL archiver objects
    """

    def __init__(self, backup_manager, name):
        """
        Base class init method.

        :param backup_manager: The backup manager
        :param name: The name of this archiver
        :return:
        """
        self.backup_manager = backup_manager
        self.server = backup_manager.server
        self.config = backup_manager.config
        self.name = name
        if self.server.use_wal_cloud_storage:
            self.storage_strategy = CloudWalStorageStrategy(
                self.backup_manager, self.server
            )
        else:
            self.storage_strategy = LocalWalStorageStrategy(
                self.backup_manager, self.server
            )
        super(WalArchiver, self).__init__()

    def receive_wal(self, reset=False):
        """
        Manage reception of WAL files. Does nothing by default.
        Some archiver classes, like the StreamingWalArchiver, have a full
        implementation.

        :param bool reset: When set, resets the status of receive-wal
        :raise ArchiverFailure: when something goes wrong
        """

    def archive(self, verbose=True):
        """
        Archive WAL files, discarding duplicates or those that are not valid.

        :param boolean verbose: Flag for verbose output
        """
        compressor = self.backup_manager.compression_manager.get_default_compressor()
        encryption = self.backup_manager.encryption_manager.get_encryption()
        processed = 0
        header = "Processing xlog segments from %s for %s" % (
            self.name,
            self.config.name,
        )

        # Get the next batch of WAL files to be processed
        batch = self.get_next_batch()

        # Analyse the batch and properly log the information
        if batch.run_size:
            if batch.total_size > batch.run_size:
                # Batch mode enabled
                _logger.info(
                    "Found %s xlog segments from %s for %s."
                    " Archive a batch of %s segments in this run.",
                    batch.total_size,
                    self.name,
                    self.config.name,
                    batch.run_size,
                )
                header += " (batch size: %s)" % batch.run_size
            else:
                # Single run mode (traditional)
                _logger.info(
                    "Found %s xlog segments from %s for %s."
                    " Archive all segments in one run.",
                    batch.total_size,
                    self.name,
                    self.config.name,
                )
        else:
            _logger.info(
                "No xlog segments found from %s for %s.", self.name, self.config.name
            )

        # Print the header (verbose mode)
        if verbose:
            output.info(header, log=False)

        # Loop through all available WAL files
        for wal_info in batch:
            # Print the header (non verbose mode)
            if not processed and not verbose:
                output.info(header, log=False)

            processed += 1

            # Report to the user the WAL file we are archiving
            output.info("\t%s", wal_info.name, log=False)
            _logger.info(
                "Archiving segment %s of %s from %s: %s/%s",
                processed,
                batch.run_size,
                self.name,
                self.config.name,
                wal_info.name,
            )
            # Archive the WAL file
            try:
                self.storage_strategy.save(compressor, encryption, wal_info)
            except MatchingDuplicateWalFile:
                # We already have this file. Simply unlink the file.
                os.unlink(wal_info.orig_filename)
                continue
            except DuplicateWalFile:
                self.server.move_wal_file_to_errors_directory(
                    wal_info.orig_filename, wal_info.name, "duplicate"
                )
                output.info(
                    "\tError: %s is already present in server %s. "
                    "File moved to errors directory.",
                    wal_info.name,
                    self.config.name,
                )
                continue
            except AbortedRetryHookScript as e:
                _logger.warning(
                    "Archiving of %s/%s aborted by "
                    "pre_archive_retry_script."
                    "Reason: %s" % (self.config.name, wal_info.name, e)
                )
                return

        if processed:
            if batch.total_size > batch.run_size:
                _logger.debug(
                    "Batch size reached (%s) - Exit %s process for %s",
                    batch.batch_size,
                    self.name,
                    self.config.name,
                )

            _logger.debug(
                "Archived %s out of %s xlog segments from %s for %s",
                processed,
                batch.total_size,
                self.name,
                self.config.name,
            )
        elif verbose:
            output.info("\tno file found", log=False)

        if batch.errors:
            output.info(
                "Some unknown objects have been found while "
                "processing xlog segments for %s. "
                "Objects moved to errors directory:",
                self.config.name,
                log=False,
            )
            # Log unexpected files
            _logger.warning(
                "Archiver is about to move %s unexpected file(s) "
                "to errors directory for %s from %s",
                len(batch.errors),
                self.config.name,
                self.name,
            )
            for error in batch.errors:
                basename = os.path.basename(error)
                output.info("\t%s", basename, log=False)
                # Print informative log line.
                _logger.warning(
                    "Moving unexpected file for %s from %s: %s",
                    self.config.name,
                    self.name,
                    basename,
                )
                self.server.move_wal_file_to_errors_directory(
                    error, basename, "unknown"
                )

    @abstractmethod
    def get_next_batch(self):
        """
        Return a WalArchiverQueue containing the WAL files to be archived.

        :rtype: WalArchiverQueue
        """

    @abstractmethod
    def check(self, check_strategy):
        """
        Perform specific checks for the archiver - invoked
        by server.check_postgres

        :param CheckStrategy check_strategy: the strategy for the management
             of the results of the various checks
        """

    @abstractmethod
    def status(self):
        """
        Set additional status info - invoked by Server.status()
        """

    @staticmethod
    def summarise_error_files(error_files):
        """
        Summarise a error files list

        :param list[str] error_files: Error files list to summarise
        :return str: A summary, None if there are no error files
        """

        if not error_files:
            return None

        # The default value for this dictionary will be 0
        counters = collections.defaultdict(int)

        # Count the file types
        for name in error_files:
            if name.endswith(".error"):
                counters["not relevant"] += 1
            elif name.endswith(".duplicate"):
                counters["duplicates"] += 1
            elif name.endswith(".unknown"):
                counters["unknown"] += 1
            else:
                counters["unknown failure"] += 1

        # Return a summary list of the form: "item a: 2, item b: 5"
        return ", ".join("%s: %s" % entry for entry in counters.items())


class FileWalArchiver(WalArchiver):
    """
    Manager of file-based WAL archiving operations (aka 'log shipping').
    """

    def __init__(self, backup_manager):
        super(FileWalArchiver, self).__init__(backup_manager, "file archival")

    def fetch_remote_status(self):
        """
        Returns the status of the FileWalArchiver.

        This method does not raise any exception in case of errors,
        but set the missing values to None in the resulting dictionary.

        :rtype: dict[str, None|str]
        """
        result = dict.fromkeys(["archive_mode", "archive_command"], None)
        postgres = self.server.postgres
        # If Postgres is not available we cannot detect anything
        if not postgres:
            return result
        # Query the database for 'archive_mode' and 'archive_command'
        result["archive_mode"] = postgres.get_setting("archive_mode")
        result["archive_command"] = postgres.get_setting("archive_command")

        # Add pg_stat_archiver statistics if the view is supported
        pg_stat_archiver = postgres.get_archiver_stats()
        if pg_stat_archiver is not None:
            result.update(pg_stat_archiver)
        return result

    def get_next_batch(self):
        """
        Returns the next batch of WAL files that have been archived through
        a PostgreSQL's 'archive_command' (in the 'incoming' directory)

        :return: WalArchiverQueue: list of WAL files
        """
        # Get the batch size from configuration (0 = unlimited)
        batch_size = self.config.archiver_batch_size
        # List and sort all files in the incoming directory
        # IMPORTANT: the list is sorted, and this allows us to know that the
        # WAL stream we have is monotonically increasing. That allows us to
        # verify that a backup has all the WALs required for the restore.
        file_names = glob(os.path.join(self.config.incoming_wals_directory, "*"))
        file_names.sort()
        total_size = len(file_names)
        # If batch size is set, limit the number of files to the batch size. The idea
        # is to avoid the overhead of creating several unused WalFileInfo objects. See
        # the note in the WalArchiverQueue class.
        if batch_size > 0:
            file_names = file_names[:batch_size]

        # Process anything that looks like a valid WAL file. Anything
        # else is treated like an error/anomaly
        files = []
        errors = []
        for file_name in file_names:
            # Ignore temporary files
            if file_name.endswith(".tmp"):
                continue
            if xlog.is_any_xlog_file(file_name) and os.path.isfile(file_name):
                files.append(file_name)
            else:
                errors.append(file_name)

        # Build the list of WalFileInfo
        wal_files = [
            WalFileInfo.from_file(
                filename=f,
                compression_manager=self.backup_manager.compression_manager,
                unidentified_compression=None,
                encryption_manager=self.backup_manager.encryption_manager,
                # We don't try to guess if the WAL is encrypted here. If the user sets
                # an archive_command which encrypts the WAL file, it's up to the user to
                # decrypt them later, and Barman won't do anything about it. We still
                # attempt to guess compression, though, because that's been the behavior
                # of Barman for a long time.
                encryption=None,
            )
            for f in files
        ]
        return WalArchiverQueue(
            wal_files, batch_size=batch_size, errors=errors, total_size=total_size
        )

    def check(self, check_strategy):
        """
        Perform additional checks for FileWalArchiver - invoked
        by server.check_postgres

        :param CheckStrategy check_strategy: the strategy for the management
             of the results of the various checks
        """
        check_strategy.init_check("archive_mode")
        remote_status = self.get_remote_status()
        # If archive_mode is None, there are issues connecting to PostgreSQL
        if remote_status["archive_mode"] is None:
            return
        # Check archive_mode parameter: must be on
        if remote_status["archive_mode"] in ("on", "always"):
            check_strategy.result(self.config.name, True)
        else:
            msg = "please set it to 'on'"
            if self.server.postgres.server_version >= 90500:
                msg += " or 'always'"
            check_strategy.result(self.config.name, False, hint=msg)
        check_strategy.init_check("archive_command")
        if (
            remote_status["archive_command"]
            and remote_status["archive_command"] != "(disabled)"
        ):
            check_strategy.result(self.config.name, True, check="archive_command")

            # Report if the archiving process works without issues.
            # Skip if the archive_command check fails
            # It can be None if PostgreSQL is older than 9.4
            if remote_status.get("is_archiving") is not None:
                check_strategy.result(
                    self.config.name,
                    remote_status["is_archiving"],
                    check="continuous archiving",
                )
        else:
            check_strategy.result(
                self.config.name,
                False,
                hint="please set it accordingly to documentation",
            )

    def status(self):
        """
        Set additional status info - invoked by Server.status()
        """
        # We need to get full info here from the server
        remote_status = self.server.get_remote_status()

        # If archive_mode is None, there are issues connecting to PostgreSQL
        if remote_status["archive_mode"] is None:
            return

        output.result(
            "status",
            self.config.name,
            "archive_command",
            "PostgreSQL 'archive_command' setting",
            remote_status["archive_command"]
            or "FAILED (please set it accordingly to documentation)",
        )
        last_wal = remote_status.get("last_archived_wal")
        # If PostgreSQL is >= 9.4 we have the last_archived_time
        if last_wal and remote_status.get("last_archived_time"):
            last_wal += ", at %s" % (remote_status["last_archived_time"].ctime())
        output.result(
            "status",
            self.config.name,
            "last_archived_wal",
            "Last archived WAL",
            last_wal or "No WAL segment shipped yet",
        )
        # Set output for WAL archive failures (PostgreSQL >= 9.4)
        if remote_status.get("failed_count") is not None:
            remote_fail = str(remote_status["failed_count"])
            if int(remote_status["failed_count"]) > 0:
                remote_fail += " (%s at %s)" % (
                    remote_status["last_failed_wal"],
                    remote_status["last_failed_time"].ctime(),
                )
            output.result(
                "status",
                self.config.name,
                "failed_count",
                "Failures of WAL archiver",
                remote_fail,
            )
        # Add hourly archive rate if available (PostgreSQL >= 9.4) and > 0
        if remote_status.get("current_archived_wals_per_second"):
            output.result(
                "status",
                self.config.name,
                "server_archived_wals_per_hour",
                "Server WAL archiving rate",
                "%0.2f/hour"
                % (3600 * remote_status["current_archived_wals_per_second"]),
            )


class StreamingWalArchiver(WalArchiver):
    """
    Object used for the management of streaming WAL archive operation.
    """

    def __init__(self, backup_manager):
        super(StreamingWalArchiver, self).__init__(backup_manager, "streaming")

    def fetch_remote_status(self):
        """
        Execute checks for replication-based wal archiving

        This method does not raise any exception in case of errors,
        but set the missing values to None in the resulting dictionary.

        :rtype: dict[str, None|str]
        """
        remote_status = dict.fromkeys(
            (
                "pg_receivexlog_compatible",
                "pg_receivexlog_installed",
                "pg_receivexlog_path",
                "pg_receivexlog_supports_slots",
                "pg_receivexlog_synchronous",
                "pg_receivexlog_version",
            ),
            None,
        )

        # Test pg_receivexlog existence
        version_info = PgReceiveXlog.get_version_info(self.server.path)
        if version_info["full_path"]:
            remote_status["pg_receivexlog_installed"] = True
            remote_status["pg_receivexlog_path"] = version_info["full_path"]
            remote_status["pg_receivexlog_version"] = version_info["full_version"]
            pgreceivexlog_version = version_info["major_version"]
        else:
            remote_status["pg_receivexlog_installed"] = False
            return remote_status

        # Retrieve the PostgreSQL version
        pg_version = None
        if self.server.streaming is not None:
            pg_version = self.server.streaming.server_major_version

        # If one of the version is unknown we cannot compare them
        if pgreceivexlog_version is None or pg_version is None:
            return remote_status

        # pg_version is not None so transform into a Version object
        # for easier comparison between versions
        pg_version = Version(pg_version)

        # Set conservative default values (False) for modern features
        remote_status["pg_receivexlog_compatible"] = False
        remote_status["pg_receivexlog_supports_slots"] = False
        remote_status["pg_receivexlog_synchronous"] = False

        # pg_receivexlog 9.2 is compatible only with PostgreSQL 9.2.
        if "9.2" == pg_version == pgreceivexlog_version:
            remote_status["pg_receivexlog_compatible"] = True

        # other versions are compatible with lesser versions of PostgreSQL
        # WARNING: The development versions of `pg_receivexlog` are considered
        # higher than the stable versions here, but this is not an issue
        # because it accepts everything that is less than
        # the `pg_receivexlog` version(e.g. '9.6' is less than '9.6devel')
        elif "9.2" < pg_version <= pgreceivexlog_version:
            # At least PostgreSQL 9.3 is required here
            remote_status["pg_receivexlog_compatible"] = True

            # replication slots are supported starting from version 9.4
            if "9.4" <= pg_version <= pgreceivexlog_version:
                remote_status["pg_receivexlog_supports_slots"] = True

            # Synchronous WAL streaming requires replication slots
            # and pg_receivexlog >= 9.5
            if "9.4" <= pg_version and "9.5" <= pgreceivexlog_version:
                remote_status["pg_receivexlog_synchronous"] = self._is_synchronous()

        return remote_status

    def receive_wal(self, reset=False):
        """
        Creates a PgReceiveXlog object and issues the pg_receivexlog command
        for a specific server

        :param bool reset: When set reset the status of receive-wal
        :raise ArchiverFailure: when something goes wrong
        """
        # Ensure the presence of the destination directory
        mkpath(self.config.streaming_wals_directory)

        # Execute basic sanity checks on PostgreSQL connection
        streaming_status = self.server.streaming.get_remote_status()
        if streaming_status["streaming_supported"] is None:
            raise ArchiverFailure(
                "failed opening the PostgreSQL streaming connection "
                "for server %s" % (self.config.name)
            )
        elif not streaming_status["streaming_supported"]:
            raise ArchiverFailure(
                "PostgreSQL version too old (%s < 9.2)"
                % self.server.streaming.server_txt_version
            )
        # Execute basic sanity checks on pg_receivexlog
        command = "pg_receivewal"
        if self.server.streaming.server_version < 100000:
            command = "pg_receivexlog"
        remote_status = self.get_remote_status()
        if not remote_status["pg_receivexlog_installed"]:
            raise ArchiverFailure("%s not present in $PATH" % command)
        if not remote_status["pg_receivexlog_compatible"]:
            raise ArchiverFailure(
                "%s version not compatible with PostgreSQL server version" % command
            )

        # Execute sanity check on replication slot usage
        postgres_status = self.server.postgres.get_remote_status()
        if self.config.slot_name:
            # Check if slots are supported
            if not remote_status["pg_receivexlog_supports_slots"]:
                raise ArchiverFailure(
                    "Physical replication slot not supported by %s "
                    "(9.4 or higher is required)"
                    % self.server.streaming.server_txt_version
                )
            # Check if the required slot exists
            if postgres_status["replication_slot"] is None:
                if self.config.create_slot == "auto":
                    if not reset:
                        output.info(
                            "Creating replication slot '%s'", self.config.slot_name
                        )
                        self.server.create_physical_repslot()
                else:
                    raise ArchiverFailure(
                        "replication slot '%s' doesn't exist. "
                        "Please execute "
                        "'barman receive-wal --create-slot %s'"
                        % (self.config.slot_name, self.config.name)
                    )
            # Check if the required slot is available
            elif postgres_status["replication_slot"].active:
                raise ArchiverFailure(
                    "replication slot '%s' is already in use" % (self.config.slot_name,)
                )

        # Check if is a reset request
        if reset:
            self._reset_streaming_status(postgres_status, streaming_status)
            return

        # Check the size of the .partial WAL file and truncate it if needed
        self._truncate_partial_file_if_needed(postgres_status["xlog_segment_size"])

        # Make sure we are not wasting precious PostgreSQL resources
        self.server.close()

        _logger.info("Activating WAL archiving through streaming protocol")
        try:
            output_handler = PgReceiveXlog.make_output_handler(self.config.name + ": ")
            receive = PgReceiveXlog(
                connection=self.server.streaming,
                destination=self.config.streaming_wals_directory,
                command=remote_status["pg_receivexlog_path"],
                version=remote_status["pg_receivexlog_version"],
                app_name=self.config.streaming_archiver_name,
                path=self.server.path,
                slot_name=self.config.slot_name,
                synchronous=remote_status["pg_receivexlog_synchronous"],
                out_handler=output_handler,
                err_handler=output_handler,
            )
            # Finally execute the pg_receivexlog process
            receive.execute()
        except CommandFailedException as e:
            # Retrieve the return code from the exception
            ret_code = e.args[0]["ret"]
            if ret_code < 0:
                # If the return code is negative, then pg_receivexlog
                # was terminated by a signal
                msg = "%s terminated by signal: %s" % (command, abs(ret_code))
            else:
                # Otherwise terminated with an error
                msg = "%s terminated with error code: %s" % (command, ret_code)

            raise ArchiverFailure(msg)
        except KeyboardInterrupt:
            # This is a normal termination, so there is nothing to do beside
            # informing the user.
            output.info("SIGINT received. Terminate gracefully.")

    def _reset_streaming_status(self, postgres_status, streaming_status):
        """
        Reset the status of receive-wal by removing the .partial file that
        is marking the current position and creating one that is current with
        the PostgreSQL insert location
        """
        current_wal = xlog.location_to_xlogfile_name_offset(
            postgres_status["current_lsn"],
            streaming_status["timeline"],
            postgres_status["xlog_segment_size"],
        )["file_name"]
        restart_wal = current_wal
        if (
            postgres_status["replication_slot"]
            and postgres_status["replication_slot"].restart_lsn
        ):
            restart_wal = xlog.location_to_xlogfile_name_offset(
                postgres_status["replication_slot"].restart_lsn,
                streaming_status["timeline"],
                postgres_status["xlog_segment_size"],
            )["file_name"]
        restart_path = os.path.join(self.config.streaming_wals_directory, restart_wal)
        restart_partial_path = restart_path + ".partial"
        wal_files = sorted(
            glob(os.path.join(self.config.streaming_wals_directory, "*")), reverse=True
        )

        # Pick the newer file
        last = None
        for last in wal_files:
            if xlog.is_wal_file(last) or xlog.is_partial_file(last):
                break

        # Check if the status is already up-to-date
        if not last or last == restart_partial_path or last == restart_path:
            output.info("Nothing to do. Position of receive-wal is aligned.")
            return

        if os.path.basename(last) > current_wal:
            output.error(
                "The receive-wal position is ahead of PostgreSQL "
                "current WAL lsn (%s > %s)",
                os.path.basename(last),
                postgres_status["current_xlog"],
            )
            return

        output.info("Resetting receive-wal directory status")
        if xlog.is_partial_file(last):
            output.info("Removing status file %s" % last)
            os.unlink(last)
        output.info("Creating status file %s" % restart_partial_path)
        open(restart_partial_path, "w").close()

    def _truncate_partial_file_if_needed(self, xlog_segment_size):
        """
        Truncate .partial WAL file if size is not 0 or xlog_segment_size

        :param int xlog_segment_size:
        """
        # Retrieve the partial list (only one is expected)
        partial_files = glob(
            os.path.join(self.config.streaming_wals_directory, "*.partial")
        )

        # Take the last partial file, ignoring wrongly formatted file names
        last_partial = None
        for partial in partial_files:
            if not is_partial_file(partial):
                continue
            if not last_partial or partial > last_partial:
                last_partial = partial

        # Skip further work if there is no good partial file
        if not last_partial:
            return

        # If size is either 0 or wal_segment_size everything is fine...
        partial_size = os.path.getsize(last_partial)
        if partial_size == 0 or partial_size == xlog_segment_size:
            return

        # otherwise truncate the file to be empty. This is safe because
        # pg_receivewal pads the file to the full size before start writing.
        output.info(
            "Truncating partial file %s that has wrong size %s "
            "while %s was expected." % (last_partial, partial_size, xlog_segment_size)
        )
        open(last_partial, "wb").close()

    def get_next_batch(self):
        """
        Returns the next batch of WAL files that have been archived via
        streaming replication (in the 'streaming' directory)

        This method always leaves one file in the "streaming" directory,
        because the 'pg_receivexlog' process needs at least one file to
        detect the current streaming position after a restart.

        :return: WalArchiverQueue: list of WAL files
        """
        # Get the batch size from configuration (0 = unlimited)
        batch_size = self.config.streaming_archiver_batch_size
        # List and sort all files in the incoming directory.
        # IMPORTANT: the list is sorted, and this allows us to know that the
        # WAL stream we have is monotonically increasing. That allows us to
        # verify that a backup has all the WALs required for the restore.
        file_names = glob(os.path.join(self.config.streaming_wals_directory, "*"))
        file_names.sort()
        total_size = len(file_names)
        # If batch size is set, limit the number of files to the batch size. The idea
        # is to avoid the overhead of creating several unused WalFileInfo objects. See
        # the note in the WalArchiverQueue class.
        if batch_size > 0:
            file_names = file_names[:batch_size]

        # Process anything that looks like a valid WAL file,
        # including partial ones and history files.
        # Anything else is treated like an error/anomaly
        files = []
        skip = []
        errors = []
        for file_name in file_names:
            # Ignore temporary files
            if file_name.endswith(".tmp"):
                continue
            # If the file doesn't exist, it has been renamed/removed while
            # we were reading the directory. Ignore it.
            if not os.path.exists(file_name):
                continue
            if not os.path.isfile(file_name):
                errors.append(file_name)
            elif xlog.is_partial_file(file_name):
                skip.append(file_name)
            elif xlog.is_any_xlog_file(file_name):
                files.append(file_name)
            else:
                errors.append(file_name)
        # In case of more than a partial file, keep the last
        # and treat the rest as normal files
        if len(skip) > 1:
            partials = skip[:-1]
            _logger.info(
                "Archiving partial files for server %s: %s"
                % (self.config.name, ", ".join([os.path.basename(f) for f in partials]))
            )
            files.extend(partials)
            skip = skip[-1:]

        # Keep the last full WAL file in case no partial file is present
        elif len(skip) == 0 and files:
            skip.append(files.pop())

        # Build the list of WalFileInfo
        wal_files = [
            WalFileInfo.from_file(
                filename=f,
                compression_manager=self.backup_manager.compression_manager,
                encryption_manager=self.backup_manager.encryption_manager,
                unidentified_compression=None,
                # WAL files received through pg_receivewal are surely not encrypted nor
                # compressed, so we avoid the overhead of trying to guess such
                # algorithms.
                compression=None,
                encryption=None,
            )
            for f in files
        ]
        return WalArchiverQueue(
            wal_files,
            batch_size=batch_size,
            errors=errors,
            skip=skip,
            total_size=total_size,
        )

    def check(self, check_strategy):
        """
        Perform additional checks for StreamingWalArchiver - invoked
        by server.check_postgres

        :param CheckStrategy check_strategy: the strategy for the management
             of the results of the various checks
        """

        check_strategy.init_check("pg_receivexlog")
        # Check the version of pg_receivexlog
        remote_status = self.get_remote_status()
        check_strategy.result(
            self.config.name, remote_status["pg_receivexlog_installed"]
        )
        hint = None
        check_strategy.init_check("pg_receivexlog compatible")
        if not remote_status["pg_receivexlog_compatible"]:
            pg_version = "Unknown"
            if self.server.streaming is not None:
                pg_version = self.server.streaming.server_txt_version
            hint = "PostgreSQL version: %s, pg_receivexlog version: %s" % (
                pg_version,
                remote_status["pg_receivexlog_version"],
            )
        check_strategy.result(
            self.config.name, remote_status["pg_receivexlog_compatible"], hint=hint
        )

        # Check if pg_receivexlog is running, by retrieving a list
        # of running 'receive-wal' processes from the process manager.
        receiver_list = self.server.process_manager.list("receive-wal")

        # If there's at least one 'receive-wal' process running for this
        # server, the test is passed
        check_strategy.init_check("receive-wal running")
        if receiver_list:
            check_strategy.result(self.config.name, True)
        else:
            check_strategy.result(
                self.config.name, False, hint="See the Barman log file for more details"
            )

    def _is_synchronous(self):
        """
        Check if receive-wal process is eligible for synchronous replication

        The receive-wal process is eligible for synchronous replication
        if `synchronous_standby_names` is configured and contains
        the value of `streaming_archiver_name`

        :rtype: bool
        """
        # Nothing to do if postgres connection is not working
        postgres = self.server.postgres
        if postgres is None or postgres.server_txt_version is None:
            return None

        # Check if synchronous WAL streaming can be enabled
        # by peeking 'synchronous_standby_names'
        postgres_status = postgres.get_remote_status()
        syncnames = postgres_status["synchronous_standby_names"]
        _logger.debug(
            "Look for '%s' in 'synchronous_standby_names': %s",
            self.config.streaming_archiver_name,
            syncnames,
        )
        # The receive-wal process is eligible for synchronous replication
        # if `synchronous_standby_names` is configured and contains
        # the value of `streaming_archiver_name`
        streaming_archiver_name = self.config.streaming_archiver_name
        synchronous = syncnames and (
            "*" in syncnames or streaming_archiver_name in syncnames
        )
        _logger.debug(
            "Synchronous WAL streaming for %s: %s", streaming_archiver_name, synchronous
        )
        return synchronous

    def status(self):
        """
        Set additional status info - invoked by Server.status()
        """
        # TODO: Add status information for WAL streaming



# -*- coding: utf-8 -*-
# © Copyright EnterpriseDB UK Limited 2018-2025
#
# This file is part of Barman.
#
# Barman is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Barman is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Barman.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os
from contextlib import closing
from operator import attrgetter

from barman import xlog
from barman.backup import BackupManager
from barman.clients.cloud_cli import (
    CLIErrorExit,
    GeneralErrorExit,
    OperationErrorExit,
    create_argument_parser,
)
from barman.cloud import CloudBackupCatalog, configure_logging
from barman.cloud_providers import (
    get_cloud_interface,
    get_snapshot_interface_from_backup_info,
)
from barman.cloud_providers.aws_s3 import S3CloudInterface
from barman.exceptions import BadXlogPrefix, InvalidRetentionPolicy
from barman.retention_policies import RetentionPolicyFactory
from barman.utils import check_non_negative, force_str

_logger = logging.getLogger(__name__)


def _get_files_for_backup(catalog, backup_info):
    backup_files = []
    # Sort the files by OID so that we always get a stable order. The PGDATA dir
    # has no OID so we use a -1 for sorting purposes, such that it always sorts
    # ahead of the tablespaces.
    for oid, backup_file in sorted(
        catalog.get_backup_files(backup_info, allow_missing=True).items(),
        key=lambda x: x[0] if x[0] else -1,
    ):
        key = oid or "PGDATA"
        for file_info in [backup_file] + sorted(
            backup_file.additional_files, key=attrgetter("path")
        ):
            # Silently skip files which could not be found - if they don't exist
            # then not being able to delete them is not an error condition here
            if file_info.path is not None:
                _logger.debug(
                    "Will delete archive for %s at %s" % (key, file_info.path)
                )
                backup_files.append(file_info.path)

    return backup_files


def _remove_wals_for_backup(
    cloud_interface,
    catalog,
    deleted_backup,
    dry_run,
    skip_wal_cleanup_if_standalone=True,
):
    """
    Remove WAL files that are no longer needed after a backup deletion.

    This function implements WAL cleanup logic for cloud backups. It identifies and
    deletes WAL files that are no longer required by any remaining backups while
    preserving WALs needed for archival standalone backups and other timelines.

    :param cloud_interface: Interface for interacting with cloud storage
    :type cloud_interface: :class:`CloudInterface`
    :param catalog: Cloud backup catalog containing backup and WAL metadata
    :type catalog: :class:`CloudBackupCatalog`
    :param deleted_backup: The backup that is being deleted
    :type deleted_backup: :class:`BackupInfo`
    :param dry_run: If ``True``, identify objects for deletion but do not delete them
    :type dry_run: :class:`bool`
    :param skip_wal_cleanup_if_standalone: If ``True``, skip WAL cleanup when deleting
        a standalone archival backup. Defaults to ``True``.
    :type skip_wal_cleanup_if_standalone: :class:`bool`

    :return: ``None``. Exits early if an error occurs during WAL listing.
    """
    # An implementation of BackupManager.remove_wal_before_backup which does not
    # use xlogdb, since xlogdb is not available to barman-cloud
    should_remove_wals, wal_ranges_to_protect = BackupManager.should_remove_wals(
        deleted_backup,
        catalog.get_backup_list(),
        keep_manager=catalog,
        skip_wal_cleanup_if_standalone=skip_wal_cleanup_if_standalone,
    )
    next_backup = BackupManager.find_next_backup_in(
        catalog.get_backup_list(), deleted_backup.backup_id
    )
    wals_to_delete = {}
    if should_remove_wals:
        # There is no previous backup or all previous backups are archival
        # standalone backups, so we can remove unused WALs (those WALs not
        # required by standalone archival backups).
        # If there is a next backup then all unused WALs up to the begin_wal
        # of the next backup can be removed.
        # If there is no next backup then there are no remaining backups,
        # because we must assume non-exclusive backups are taken, we can only
        # safely delete unused WALs up to begin_wal of the deleted backup.
        # See comments in barman.backup.BackupManager.delete_backup.
        if next_backup:
            remove_until = next_backup
        else:
            remove_until = deleted_backup
        # A WAL is only a candidate for deletion if it is on the same timeline so we
        # use BackupManager to get a set of all other timelines with backups so that
        # we can preserve all WALs on other timelines.
        timelines_to_protect = BackupManager.get_timelines_to_protect(
            remove_until=remove_until,
            deleted_backup=deleted_backup,
            available_backups=catalog.get_backup_list(),
        )
        # Identify any prefixes under which all WALs are no longer needed.
        # This is a shortcut which allows us to delete all WALs under a prefix without
        # checking each individual WAL.
        try:
            wal_prefixes = catalog.get_wal_prefixes()
        except NotImplementedError:
            # If fetching WAL prefixes isn't supported by the cloud provider then
            # the old method of checking each WAL must be used for all WALs.
            wal_prefixes = []
        deletable_prefixes = []
        for wal_prefix in wal_prefixes:
            try:
                tli_and_log = wal_prefix.split("/")[-2]
                tli, log = xlog.decode_hash_dir(tli_and_log)
            except (BadXlogPrefix, IndexError):
                # If the prefix does not appear to be a tli and log we output a warning
                # and move on to the next prefix rather than error out.
                _logger.warning(
                    "Ignoring malformed WAL object prefix: {}".format(wal_prefix)
                )
                continue
            # If this prefix contains a timeline which should be protected then we
            # cannot delete the WALS under it so advance to the next prefix.
            if tli in timelines_to_protect:
                continue

            # If the tli and log fall are inclusively between the tli and log for the
            # begin and end WAL of any protected WAL range then this prefix cannot be
            # deleted outright.
            for begin_wal, end_wal in wal_ranges_to_protect:
                begin_tli, begin_log, _ = xlog.decode_segment_name(begin_wal)
                end_tli, end_log, _ = xlog.decode_segment_name(end_wal)
                if (
                    tli >= begin_tli
                    and log >= begin_log
                    and tli <= end_tli
                    and log <= end_log
                ):
                    break
            else:
                # The prefix tli and log do not match any protected timelines or
                # protected WAL ranges so all WALs are eligible for deletion if the tli
                # is the same timeline and the log is below the begin_wal log of the
                # backup being deleted.
                until_begin_tli, until_begin_log, _ = xlog.decode_segment_name(
                    remove_until.begin_wal
                )
                if tli == until_begin_tli and log < until_begin_log:
                    # All WALs under this prefix pre-date the backup being deleted so they
                    # can be deleted in one request.
                    deletable_prefixes.append(wal_prefix)
        for wal_prefix in deletable_prefixes:
            if not dry_run:
                cloud_interface.delete_under_prefix(wal_prefix)
            else:
                print(
                    "Skipping deletion of all objects under prefix %s "
                    "due to --dry-run option" % wal_prefix
                )
        try:
            wal_paths = catalog.get_wal_paths()
        except Exception as exc:
            _logger.error(
                "Cannot clean up WALs for backup %s because an error occurred listing WALs: %s",
                deleted_backup.backup_id,
                force_str(exc),
            )
            return
        for wal_name, wal in wal_paths.items():
            # If the wal starts with a prefix we deleted then ignore it so that the
            # dry-run output is accurate
            if any(wal.startswith(prefix) for prefix in deletable_prefixes):
                continue
            if xlog.is_history_file(wal_name):
                continue
            if timelines_to_protect:
                tli, _, _ = xlog.decode_segment_name(wal_name)
                if tli in timelines_to_protect:
                    continue

            # Check if the WAL is in a protected range, required by an archival
            # standalone backup - so do not delete it
            if xlog.is_backup_file(wal_name):
                # If we have a backup file, truncate the name for the range check
                range_check_wal_name = wal_name[:24]
            else:
                range_check_wal_name = wal_name
            if any(
                range_check_wal_name >= begin_wal and range_check_wal_name <= end_wal
                for begin_wal, end_wal in wal_ranges_to_protect
            ):
                continue

            if wal_name < remove_until.begin_wal:
                wals_to_delete[wal_name] = wal
    # Explicitly sort because dicts are not ordered in python < 3.6
    wal_paths_to_delete = sorted(wals_to_delete.values())
    if len(wal_paths_to_delete) > 0:
        if not dry_run:
            try:
                cloud_interface.delete_objects(wal_paths_to_delete)
            except Exception as exc:
                _logger.error(
                    "Could not delete the following WALs for backup %s: %s, Reason: %s",
                    deleted_backup.backup_id,
                    wal_paths_to_delete,
                    force_str(exc),
                )
                # Return early so that we leave the WALs in the local cache so they
                # can be cleaned up should there be a subsequent backup deletion.
                return
        else:
            print(
                "Skipping deletion of objects %s due to --dry-run option"
                % wal_paths_to_delete
            )
        for wal_name in wals_to_delete.keys():
            catalog.remove_wal_from_cache(wal_name)


def _delete_backup(
    cloud_interface,
    catalog,
    backup_id,
    config,
    skip_wal_cleanup_if_standalone=True,
):
    backup_info = catalog.get_backup_info(backup_id)
    if not backup_info:
        _logger.warning("Backup %s does not exist", backup_id)
        return

    if backup_info.snapshots_info:
        _logger.debug(
            "Will delete the following snapshots: %s",
            ", ".join(
                snapshot.identifier for snapshot in backup_info.snapshots_info.snapshots
            ),
        )
        if not config.dry_run:
            snapshot_interface = get_snapshot_interface_from_backup_info(
                backup_info, config
            )
            snapshot_interface.delete_snapshot_backup(backup_info)
        else:
            print("Skipping deletion of snapshots due to --dry-run option")
        # Delete the backup_label for snapshots backups as this is not stored in the
        # same format used by the non-snapshot backups.
        backup_label_path = os.path.join(
            catalog.prefix, backup_info.backup_id, "backup_label"
        )
        if not config.dry_run:
            cloud_interface.delete_objects([backup_label_path])
        else:
            print("Skipping deletion of %s due to --dry-run option" % backup_label_path)

    objects_to_delete = _get_files_for_backup(catalog, backup_info)
    backup_info_path = os.path.join(
        catalog.prefix, backup_info.backup_id, "backup.info"
    )
    _logger.debug("Will delete backup.info file at %s" % backup_info_path)
    if not config.dry_run:
        try:
            # Check Object Lock only on base backup files.
            # If these are unlocked and deletable, we proceed to delete everything.
            cloud_interface.delete_objects(
                objects_to_delete, check_locks=config.check_object_lock, atomic=True
            )
            # Do not try to delete backup.info until we have successfully deleted
            # everything else so that it is possible to retry the operation should
            # we fail to delete any backup file
            cloud_interface.delete_objects([backup_info_path])
        except Exception as exc:
            _logger.error("Could not delete backup %s: %s", backup_id, force_str(exc))
            raise OperationErrorExit()
    else:
        print(
            "Skipping deletion of objects %s due to --dry-run option"
            % (objects_to_delete + [backup_info_path])
        )

    # Remove WALs without checking locks. Since base backup is already gone,
    # the WALs no longer have value, and it's not worth the overhead of checking
    # their lock status.
    _remove_wals_for_backup(
        cloud_interface,
        catalog,
        backup_info,
        config.dry_run,
        skip_wal_cleanup_if_standalone,
    )
    # It is important that the backup is removed from the catalog after cleaning
    # up the WALs because the code in _remove_wals_for_backup depends on the
    # deleted backup existing in the backup catalog
    catalog.remove_backup_from_cache(backup_id)


def main(args=None):
    """
    The main script entry point

    :param list[str] args: the raw arguments list. When not provided
        it defaults to sys.args[1:]
    """
    config = parse_arguments(args)
    configure_logging(config)

    try:
        cloud_interface = get_cloud_interface(config)

        with closing(cloud_interface):
            # Warn if check_object_lock is used with non-S3 providers
            if config.check_object_lock and not isinstance(
                cloud_interface, S3CloudInterface
            ):
                _logger.warning(
                    "The --check-object-lock option is currently only supported for aws-s3. "
                    "Object lock checks will not be performed for cloud provider: %s",
                    config.cloud_provider,
                )

            # Do connectivity test if requested
            if config.test:
                cloud_interface.verify_cloud_connectivity_and_bucket_existence()
                raise SystemExit(0)

            catalog = CloudBackupCatalog(
                cloud_interface=cloud_interface, server_name=config.server_name
            )
            # Call catalog.get_backup_list now so we know we can read the whole catalog
            # (the results are cached so this does not result in extra calls to cloud
            # storage)
            catalog.get_backup_list()
            if len(catalog.unreadable_backups) > 0:
                _logger.error(
                    "Cannot read the following backups: %s\n"
                    "Unsafe to proceed with deletion due to failure reading backup catalog"
                    % catalog.unreadable_backups
                )
                raise OperationErrorExit()
            if config.backup_id:
                backup_id = catalog.parse_backup_id(config.backup_id)
                # Because we only care about one backup, skip the annotation cache
                # because it is only helpful when dealing with multiple backups
                if catalog.should_keep_backup(backup_id, use_cache=False):
                    _logger.error(
                        "Skipping delete of backup %s for server %s "
                        "as it has a current keep request. If you really "
                        "want to delete this backup please remove the keep "
                        "and try again.",
                        backup_id,
                        config.server_name,
                    )
                    raise OperationErrorExit()
                if config.minimum_redundancy > 0:
                    if config.minimum_redundancy >= len(catalog.get_backup_list()):
                        _logger.error(
                            "Skipping delete of backup %s for server %s "
                            "due to minimum redundancy requirements "
                            "(minimum redundancy = %s, "
                            "current redundancy = %s)",
                            backup_id,
                            config.server_name,
                            config.minimum_redundancy,
                            len(catalog.get_backup_list()),
                        )
                        raise OperationErrorExit()
                _delete_backup(cloud_interface, catalog, backup_id, config)
            elif config.retention_policy:
                try:
                    retention_policy = RetentionPolicyFactory.create(
                        "retention_policy",
                        config.retention_policy,
                        server_name=config.server_name,
                        catalog=catalog,
                        minimum_redundancy=config.minimum_redundancy,
                    )
                except InvalidRetentionPolicy as exc:
                    _logger.error(
                        "Could not create retention policy %s: %s",
                        config.retention_policy,
                        force_str(exc),
                    )
                    raise CLIErrorExit()
                # Sort to ensure that we delete the backups in ascending order, that is
                # from oldest to newest. This ensures that the relevant WALs will be cleaned
                # up after each backup is deleted.
                backups_to_delete = sorted(
                    [
                        backup_id
                        for backup_id, status in retention_policy.report().items()
                        if status == "OBSOLETE"
                    ]
                )
                for backup_id in backups_to_delete:
                    _delete_backup(
                        cloud_interface,
                        catalog,
                        backup_id,
                        config,
                        skip_wal_cleanup_if_standalone=False,
                    )
    except Exception as exc:
        _logger.error("Barman cloud backup delete exception: %s", force_str(exc))
        _logger.debug("Exception details:", exc_info=exc)
        raise GeneralErrorExit()


def parse_arguments(args=None):
    """
    Parse command line arguments

    :return: The options parsed
    """
    parser, _, _ = create_argument_parser(
        description="This script can be used to delete backups "
        "made with barman-cloud-backup command. "
        "Currently AWS S3, Azure Blob Storage and Google Cloud Storage are supported.",
    )
    delete_arguments = parser.add_mutually_exclusive_group(required=True)
    delete_arguments.add_argument(
        "-b",
        "--backup-id",
        help="Backup ID of the backup to be deleted",
    )
    parser.add_argument(
        "-m",
        "--minimum-redundancy",
        type=check_non_negative,
        help="The minimum number of backups that should always be available.",
        default=0,
    )
    delete_arguments.add_argument(
        "-r",
        "--retention-policy",
        help="If specified, delete all backups eligible for deletion according to the "
        "supplied retention policy. Syntax: REDUNDANCY value | RECOVERY WINDOW OF "
        "value {DAYS | WEEKS | MONTHS}",
    )
    parser.add_argument(
        "--dry-run",
        action="store_true",
        help="Find the objects which need to be deleted but do not delete them",
    )
    parser.add_argument(
        "--batch-size",
        dest="delete_batch_size",
        type=int,
        help="The maximum number of objects to be deleted in a single request to the "
        "cloud provider. If unset then the maximum allowed batch size for the "
        "specified cloud provider will be used (1000 for aws-s3, 256 for "
        "azure-blob-storage and 100 for google-cloud-storage).",
    )
    parser.add_argument(
        "--check-object-lock",
        action="store_true",
        help="Check for S3 Object Lock before deleting to skip deletion of locked "
        "objects. This option adds overhead as it requires a request to the object "
        "store for each object of the base backup to delete.",
    )
    return parser.parse_args(args=args)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
# © Copyright EnterpriseDB UK Limited 2018-2025
#
# This file is part of Barman.
#
# Barman is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Barman is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Barman.  If not, see <http://www.gnu.org/licenses/>.
import logging
import os
from abc import ABCMeta, abstractmethod
from contextlib import closing

from barman.clients.cloud_cli import (
    CLIErrorExit,
    GeneralErrorExit,
    OperationErrorExit,
    create_argument_parser,
)
from barman.cloud import CloudBackupCatalog, configure_logging
from barman.cloud_providers import (
    get_cloud_interface,
    get_snapshot_interface_from_backup_info,
)
from barman.exceptions import ConfigurationException
from barman.fs import UnixLocalCommand
from barman.recovery_executor import SnapshotRecoveryExecutor
from barman.utils import (
    check_tli,
    force_str,
    get_backup_id_from_target_lsn,
    get_backup_id_from_target_time,
    get_backup_id_from_target_tli,
    get_last_backup_id,
    parse_target_tli,
    with_metaclass,
)

_logger = logging.getLogger(__name__)


def _validate_config(config, backup_info):
    """
    Additional validation for config such as mutually inclusive options.

    Raises a ConfigurationException if any options are missing or incompatible.

    :param argparse.Namespace config: The backup options provided at the command line.
    :param BackupInfo backup_info: The backup info for the backup to restore
    """
    if backup_info.snapshots_info:
        if config.tablespace != []:
            raise ConfigurationException(
                "Backup %s is a snapshot backup therefore tablespace relocation rules "
                "cannot be used." % backup_info.backup_id,
            )


def main(args=None):
    """
    The main script entry point

    :param list[str] args: the raw arguments list. When not provided
        it defaults to sys.args[1:]
    """
    config = parse_arguments(args)
    configure_logging(config)

    try:
        cloud_interface = get_cloud_interface(config)

        with closing(cloud_interface):
            # Do connectivity test if requested
            if config.test:
                cloud_interface.verify_cloud_connectivity_and_bucket_existence()
                raise SystemExit(0)

            catalog = CloudBackupCatalog(cloud_interface, config.server_name)
            backup_id = None
            if config.backup_id != "auto":
                backup_id = catalog.parse_backup_id(config.backup_id)
            else:
                target_options = ["target_time", "target_lsn"]
                target_option = None
                for option in target_options:
                    target = getattr(config, option, None)
                    if target is not None:
                        target_option = option
                        break

                # "Parse" the string value to integer for `target_tli` if passed as a
                # string ("current", "latest")
                target_tli = parse_target_tli(obj=catalog, target_tli=config.target_tli)

                available_backups = catalog.get_backup_list().values()
                if target_option is None:
                    if target_tli is not None:
                        backup_id = get_backup_id_from_target_tli(
                            available_backups, target_tli
                        )
                    else:
                        backup_id = get_last_backup_id(available_backups)
                elif target_option == "target_time":
                    backup_id = get_backup_id_from_target_time(
                        available_backups, target, target_tli
                    )
                elif target_option == "target_lsn":
                    backup_id = get_backup_id_from_target_lsn(
                        available_backups, target, target_tli
                    )
                # If no candidate backup_id is found, error out.
                if backup_id is None:
                    _logger.error("Cannot find any candidate backup for recovery.")
                    raise OperationErrorExit()

            backup_info = catalog.get_backup_info(backup_id)
            _logger.info("Restoring from backup_id: %s" % backup_id)
            if not backup_info:
                _logger.error(
                    "Backup %s for server %s does not exists",
                    backup_id,
                    config.server_name,
                )
                raise OperationErrorExit()

            _validate_config(config, backup_info)

            if backup_info.snapshots_info:
                snapshot_interface = get_snapshot_interface_from_backup_info(
                    backup_info, config
                )
                snapshot_interface.validate_restore_config(config)
                downloader = CloudBackupDownloaderSnapshot(
                    cloud_interface, catalog, snapshot_interface
                )
                downloader.download_backup(
                    backup_info,
                    config.recovery_dir,
                    config.snapshot_recovery_instance,
                )
            else:
                downloader = CloudBackupDownloaderObjectStore(cloud_interface, catalog)
                downloader.download_backup(
                    backup_info,
                    config.recovery_dir,
                    tablespace_map(config.tablespace),
                )

    except KeyboardInterrupt as exc:
        _logger.error("Barman cloud restore was interrupted by the user")
        _logger.debug("Exception details:", exc_info=exc)
        raise OperationErrorExit()
    except Exception as exc:
        _logger.error("Barman cloud restore exception: %s", force_str(exc))
        _logger.debug("Exception details:", exc_info=exc)
        raise GeneralErrorExit()


def parse_arguments(args=None):
    """
    Parse command line arguments

    :return: The options parsed
    """

    parser, s3_arguments, azure_arguments = create_argument_parser(
        description="This script can be used to download a backup "
        "previously made with barman-cloud-backup command."
        "Currently AWS S3, Azure Blob Storage and Google Cloud Storage are supported.",
    )
    parser.add_argument("backup_id", help="the backup ID")
    parser.add_argument("recovery_dir", help="the path to a directory for recovery.")
    parser.add_argument(
        "--tablespace",
        help="tablespace relocation rule",
        metavar="NAME:LOCATION",
        action="append",
        default=[],
    )
    parser.add_argument(
        "--snapshot-recovery-instance",
        help="Instance where the disks recovered from the snapshots are attached",
    )
    parser.add_argument(
        "--snapshot-recovery-zone",
        help=(
            "Zone containing the instance and disks for the snapshot recovery "
            "(deprecated: replaced by --gcp-zone)"
        ),
        dest="gcp_zone",
    )
    s3_arguments.add_argument(
        "--aws-region",
        help=(
            "Name of the AWS region where the instance and disks for snapshot "
            "recovery are located"
        ),
    )
    gcs_arguments = parser.add_argument_group(
        "Extra options for google-cloud-storage cloud provider"
    )
    gcs_arguments.add_argument(
        "--gcp-zone",
        help="Zone containing the instance and disks for the snapshot recovery",
    )
    azure_arguments.add_argument(
        "--azure-resource-group",
        help="Resource group containing the instance and disks for the snapshot recovery",
    )
    parser.add_argument("--target-tli", help="target timeline", type=check_tli)
    target_args = parser.add_mutually_exclusive_group()
    target_args.add_argument("--target-lsn", help="target LSN (Log Sequence Number)")
    target_args.add_argument(
        "--target-time",
        help="target time. You can use any valid unambiguous representation. "
        'e.g: "YYYY-MM-DD HH:MM:SS.mmm"',
    )
    return parser.parse_args(args=args)


def tablespace_map(rules):
    """
    Return a mapping from tablespace names to locations built from any
    `--tablespace name:/loc/ation` rules specified.
    """
    tablespaces = {}
    for rule in rules:
        try:
            tablespaces.update([rule.split(":", 1)])
        except ValueError:
            _logger.error(
                "Invalid tablespace relocation rule '%s'\n"
                "HINT: The valid syntax for a relocation rule is "
                "NAME:LOCATION",
                rule,
            )
            raise CLIErrorExit()
    return tablespaces


class CloudBackupDownloader(with_metaclass(ABCMeta)):
    """
    Restore a backup from cloud storage.
    """

    def __init__(self, cloud_interface, catalog):
        """
        Object responsible for handling interactions with cloud storage

        :param CloudInterface cloud_interface: The interface to use to
          upload the backup
        :param str server_name: The name of the server as configured in Barman
        :param CloudBackupCatalog catalog: The cloud backup catalog
        """
        self.cloud_interface = cloud_interface
        self.catalog = catalog

    @abstractmethod
    def download_backup(self, backup_id, destination_dir):
        """
        Download a backup from cloud storage

        :param str backup_id: The backup id to restore
        :param str destination_dir: Path to the destination directory
        """


class CloudBackupDownloaderObjectStore(CloudBackupDownloader):
    """
    Cloud storage download client for an object store backup
    """

    def download_backup(self, backup_info, destination_dir, tablespaces):
        """
        Download a backup from cloud storage

        :param BackupInfo backup_info: The backup info for the backup to restore
        :param str destination_dir: Path to the destination directory
        """
        # Validate the destination directory before starting recovery
        if os.path.exists(destination_dir) and os.listdir(destination_dir):
            _logger.error(
                "Destination %s already exists and it is not empty", destination_dir
            )
            raise OperationErrorExit()

        backup_files = self.catalog.get_backup_files(backup_info)

        # We must download and restore a bunch of .tar files that contain PGDATA
        # and each tablespace. First, we determine a target directory to extract
        # each tar file into and record these in copy_jobs. For each tablespace,
        # the location may be overridden by `--tablespace name:/new/location` on
        # the command-line; and we must also add an entry to link_jobs to create
        # a symlink from $PGDATA/pg_tblspc/oid to the correct location after the
        # downloads.

        copy_jobs = []
        link_jobs = []
        for oid in backup_files:
            file_info = backup_files[oid]
            # PGDATA is restored where requested (destination_dir)
            if oid is None:
                target_dir = destination_dir
            else:
                for tblspc in backup_info.tablespaces:
                    if oid == tblspc.oid:
                        target_dir = tblspc.location
                        if tblspc.name in tablespaces:
                            target_dir = os.path.realpath(tablespaces[tblspc.name])
                        _logger.debug(
                            "Tablespace %s (oid=%s) will be located at %s",
                            tblspc.name,
                            oid,
                            target_dir,
                        )
                        link_jobs.append(
                            ["%s/pg_tblspc/%s" % (destination_dir, oid), target_dir]
                        )
                        break
                else:
                    raise AssertionError(
                        "The backup file oid '%s' must be present "
                        "in backupinfo.tablespaces list"
                    )

            # Validate the destination directory before starting recovery
            if os.path.exists(target_dir) and os.listdir(target_dir):
                _logger.error(
                    "Destination %s already exists and it is not empty", target_dir
                )
                raise OperationErrorExit()
            copy_jobs.append([file_info, target_dir])
            for additional_file in file_info.additional_files:
                copy_jobs.append([additional_file, target_dir])

        # Now it's time to download the files
        for file_info, target_dir in copy_jobs:
            # Download the file
            _logger.debug(
                "Extracting %s to %s (%s)",
                file_info.path,
                target_dir,
                (
                    "decompressing " + file_info.compression
                    if file_info.compression
                    else "no compression"
                ),
            )
            self.cloud_interface.extract_tar(file_info.path, target_dir)

        for link, target in link_jobs:
            os.symlink(target, link)

        # If we did not restore the pg_wal directory from one of the uploaded
        # backup files, we must recreate it here. (If pg_wal was originally a
        # symlink, it would not have been uploaded.)

        wal_path = os.path.join(destination_dir, backup_info.wal_directory())
        if not os.path.exists(wal_path):
            os.mkdir(wal_path)


class CloudBackupDownloaderSnapshot(CloudBackupDownloader):
    """A minimal downloader for cloud backups which just retrieves the backup label."""

    def __init__(self, cloud_interface, catalog, snapshot_interface):
        """
        Object responsible for handling interactions with cloud storage

        :param CloudInterface cloud_interface: The interface to use to
          upload the backup
        :param str server_name: The name of the server as configured in Barman
        :param CloudBackupCatalog catalog: The cloud backup catalog
        :param CloudSnapshotInterface snapshot_interface: Interface for managing
            snapshots via a cloud provider API.
        """
        super(CloudBackupDownloaderSnapshot, self).__init__(cloud_interface, catalog)
        self.snapshot_interface = snapshot_interface

    def download_backup(
        self,
        backup_info,
        destination_dir,
        recovery_instance,
    ):
        """
        Download a backup from cloud storage

        :param BackupInfo backup_info: The backup info for the backup to restore
        :param str destination_dir: Path to the destination directory
        :param str recovery_instance: The name of the VM instance to which the disks
            cloned from the backup snapshots are attached.
        """
        attached_volumes = SnapshotRecoveryExecutor.get_attached_volumes_for_backup(
            self.snapshot_interface,
            backup_info,
            recovery_instance,
        )
        cmd = UnixLocalCommand()
        SnapshotRecoveryExecutor.check_mount_points(backup_info, attached_volumes, cmd)
        SnapshotRecoveryExecutor.check_recovery_dir_exists(destination_dir, cmd)

        # If the target directory does not exist then we will fail here because
        # it tells us the snapshot has not been restored.
        return self.cloud_interface.download_file(
            "/".join((self.catalog.prefix, backup_info.backup_id, "backup_label")),
            os.path.join(destination_dir, "backup_label"),
            decompress=None,
        )


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
# © Copyright EnterpriseDB UK Limited 2011-2025
#
# This file is part of Barman.
#
# Barman is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Barman is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Barman.  If not, see <http://www.gnu.org/licenses/>.

"""
This module is responsible to manage the compression features of Barman
"""

import binascii
import bz2
import gzip
import logging
import lzma
import shutil
from abc import ABCMeta, abstractmethod, abstractproperty
from contextlib import closing
from distutils.version import LooseVersion as Version
from io import BytesIO
from types import SimpleNamespace

from barman.command_wrappers import Command
from barman.exceptions import (
    CommandFailedException,
    CompressionException,
    CompressionIncompatibility,
    FileNotFoundException,
)
from barman.fs import unix_command_factory
from barman.utils import force_str, with_metaclass

_logger = logging.getLogger(__name__)


class CompressionManager(object):
    def __init__(self, config, path):
        """

        :param config: barman.config.ServerConfig
        :param path: str
        """
        self.config = config
        self.path = path
        self.unidentified_compression = None

        if self.config.compression == "custom":
            # If Barman is set to use the custom compression and no magic is
            # configured, it assumes that every unidentified file is custom
            # compressed.
            if self.config.custom_compression_magic is None:
                self.unidentified_compression = self.config.compression
            # If custom_compression_magic is set then we should not assume
            # unidentified files are custom compressed and should rely on the
            # magic for identification instead.
            elif isinstance(config.custom_compression_magic, str):
                # Since we know the custom compression magic we can now add it
                # to the class property.
                compression_registry["custom"].MAGIC = binascii.unhexlify(
                    config.custom_compression_magic[2:]
                )

        # Set the longest string needed to identify a compression schema.
        # This happens at instantiation time because we need to include the
        # custom_compression_magic from the config (if set).
        self.MAGIC_MAX_LENGTH = max(
            len(x.MAGIC or "") for x in compression_registry.values()
        )

    def check(self, compression=None):
        """
        This method returns True if the compression specified in the
        configuration file is present in the register, otherwise False
        """
        if not compression:
            compression = self.config.compression
        if compression not in compression_registry:
            return False
        return True

    def get_default_compressor(self):
        """
        Returns a new default compressor instance
        """
        return self.get_compressor(self.config.compression)

    def get_compressor(self, compression):
        """
        Returns a new compressor instance

        :param str compression: Compression name or none
        """
        # Check if the requested compression mechanism is allowed
        if compression and self.check(compression):
            return compression_registry[compression](
                config=self.config, compression=compression, path=self.path
            )
        return None

    def identify_compression(self, file_or_fileobj):
        """
        Try to guess the compression algorithm of the given file or file-like object.

        .. note::
            For compression methods which share the same magic number, such as ``gzip``
            and ``pigz``, the first one alphabetically in :data:`compression_registry`
            will be returned.

        :param str|file file_or_fileobj: The file path or a file-like object to check.
        :rtype: str
        """
        if not hasattr(file_or_fileobj, "read"):
            with open(file_or_fileobj, "rb") as fileobj:
                file_start = fileobj.read(self.MAGIC_MAX_LENGTH)
        else:
            file_start = file_or_fileobj.read(self.MAGIC_MAX_LENGTH)
            file_or_fileobj.seek(0)

        for file_type, cls in sorted(compression_registry.items()):
            if cls.validate(file_start):
                return file_type
        return None


class Compressor(with_metaclass(ABCMeta, object)):
    """
    Base class for all the compressors

    :cvar MAGIC: Magic bytes used to identify the compression format
    :cvar LEVEL_MIN: Minimum compression level supported by the compression algorithm
    :cvar LEVEL_MAX: Maximum compression level supported by the compression algorithm
    :cvar LEVEL_LOW: Mapped level to ``low``
    :cvar LEVEL_MEDIUM: Mapped level to ``medium``
    :cvar LEVEL_HIGH: Mapped level to ``high``
    """

    MAGIC = None
    LEVEL_MAX = None
    LEVEL_MIN = None
    LEVEL_LOW = None
    LEVEL_MEDIUM = None
    LEVEL_HIGH = None

    def __init__(self, config, compression, path=None):
        """

        :param config: barman.config.ServerConfig
        :param compression: str compression name
        :param path: str|None
        """
        self.config = config
        self.compression = compression
        self.path = path
        if isinstance(config.compression_level, int):
            if self.LEVEL_MAX is not None and config.compression_level > self.LEVEL_MAX:
                _logger.debug(
                    "Compression level %s out of range for %s, using %s instead"
                    % (config.compression_level, config.compression, self.LEVEL_MAX)
                )
                self.level = self.LEVEL_MAX
            elif (
                self.LEVEL_MIN is not None and config.compression_level < self.LEVEL_MIN
            ):
                _logger.debug(
                    "Compression level %s out of range for %s, using %s instead"
                    % (config.compression_level, config.compression, self.LEVEL_MIN)
                )
                self.level = self.LEVEL_MIN
            else:
                self.level = config.compression_level
        elif config.compression_level == "low":
            self.level = self.LEVEL_LOW
        elif config.compression_level == "high":
            self.level = self.LEVEL_HIGH
        else:
            self.level = self.LEVEL_MEDIUM

    @classmethod
    def validate(cls, file_start):
        """
        Guess if the first bytes of a file are compatible with the compression
        implemented by this class

        :param file_start: a binary string representing the first few
            bytes of a file
        :rtype: bool
        """
        return cls.MAGIC and file_start.startswith(cls.MAGIC)

    @abstractmethod
    def compress(self, src, dst):
        """
        Abstract Method for compression method

        :param str src: source file path
        :param str dst: destination file path
        """

    @abstractmethod
    def decompress(self, src, dst):
        """
        Abstract method for decompression method

        :param str src: source file path
        :param str dst: destination file path
        """


class CommandCompressor(Compressor):
    """
    Base class for compressors built on external commands
    """

    def __init__(self, config, compression, path=None):
        """
        :param config: barman.config.ServerConfig
        :param compression: str compression name
        :param path: str|None
        """
        super(CommandCompressor, self).__init__(config, compression, path)

        self._compress = None
        self._decompress = None

    def compress(self, src, dst):
        """
        Compress using the specific command defined in the subclass

        :param src: source file to compress
        :param dst: destination of the decompression
        """
        return self._compress(src, dst)

    def decompress(self, src, dst):
        """
        Decompress using the specific command defined in the subclass

        :param src: source file to decompress
        :param dst: destination of the decompression
        """
        return self._decompress(src, dst)

    def _build_command(self, pipe_command):
        """
        Build the command string and create the actual Command object

        :param pipe_command: the command used to compress/decompress
        :rtype: Command
        """
        command = "barman_command(){ "
        command += pipe_command
        command += ' > "$2" < "$1"'
        command += ";}; barman_command"
        return Command(command, shell=True, check=True, path=self.path)


class InternalCompressor(Compressor):
    """
    Base class for compressors built on python libraries

    :cvar EXTENSION: File extension for the compressed output (e.g. ``".gz"``). Must be
        set by each concrete subclass.
    """

    EXTENSION = None

    def compress(self, src, dst):
        """
        Compress using the object defined in the subclass

        :param src: source file to compress
        :param dst: destination of the decompression
        """
        try:
            with open(src, "rb") as istream:
                with closing(self._compressor(dst)) as ostream:
                    shutil.copyfileobj(istream, ostream)
        except Exception as e:
            # you won't get more information from the compressors anyway
            raise CommandFailedException(dict(ret=None, err=force_str(e), out=None))
        return 0

    def decompress(self, src, dst):
        """
        Decompress using the object defined in the subclass

        :param src: source file to decompress
        :param dst: destination of the decompression
        """
        try:
            with closing(self._decompressor(src)) as istream:
                with open(dst, "wb") as ostream:
                    shutil.copyfileobj(istream, ostream)
        except Exception as e:
            # you won't get more information from the compressors anyway
            raise CommandFailedException(dict(ret=None, err=force_str(e), out=None))
        return 0

    @abstractmethod
    def _decompressor(self, src):
        """
        Abstract decompressor factory method

        :param src: source file path
        :return: a file-like readable decompressor object
        """

    @abstractmethod
    def _compressor(self, dst):
        """
        Abstract compressor factory method

        :param dst: destination file path
        :return: a file-like writable compressor object
        """

    @abstractmethod
    def compress_in_mem(self, fileobj):
        """
        Compresses the given file-object in memory

        :param fileobj: source file-object to be compressed
        :return: a compressed file-object

        .. note::
            When implementing this method, the compressed file-object position must be
            set to ``0`` before returning it, as it is likely to be read again afterwards.
        """

    @abstractmethod
    def decompress_in_mem(self, fileobj):
        """
        Decompresses the given file-object in memory

        :param fileobj: source file-object to be decompressed
        :return: a decompressed file-object
        """

    def decompress_to_fileobj(self, src_fileobj, dest_fileobj):
        """
        Decompresses the given file-object on the especified file-object

        :param src_fileobj: source file-object to be decompressed
        :param dest_fileobj: destination file-object to have the decompressed content
        """
        decompressed_fileobj = self.decompress_in_mem(src_fileobj)
        shutil.copyfileobj(decompressed_fileobj, dest_fileobj)


class PyGZipCompressor(InternalCompressor):
    """
    Predefined compressor that uses GZip Python libraries
    """

    EXTENSION = ".gz"
    MAGIC = b"\x1f\x8b\x08"
    LEVEL_MIN = 1
    LEVEL_MAX = 9
    LEVEL_LOW = 1
    LEVEL_MEDIUM = 6
    LEVEL_HIGH = 9

    def __init__(self, config, compression, path=None):
        """

        :param config: barman.config.ServerConfig
        :param compression: str compression name
        :param path: str|None
        """
        super(PyGZipCompressor, self).__init__(config, compression, path)

    def _compressor(self, name):
        return gzip.GzipFile(name, mode="wb", compresslevel=self.level)

    def _decompressor(self, name):
        return gzip.GzipFile(name, mode="rb")

    def compress_in_mem(self, fileobj):
        in_mem_gzip = BytesIO()
        with gzip.GzipFile(
            fileobj=in_mem_gzip, mode="wb", compresslevel=self.level
        ) as gz:
            shutil.copyfileobj(fileobj, gz)
        in_mem_gzip.seek(0)
        return in_mem_gzip

    def decompress_in_mem(self, fileobj):
        return gzip.GzipFile(fileobj=fileobj, mode="rb")


class PigzCompressor(CommandCompressor):
    """
    Predefined compressor with Pigz

    Note that pigz on-disk is the same as gzip, so the MAGIC value of this
    class is the same
    """

    MAGIC = b"\x1f\x8b\x08"
    LEVEL_MIN = 1
    LEVEL_MAX = 9
    LEVEL_LOW = 1
    LEVEL_MEDIUM = 6
    LEVEL_HIGH = 9

    def __init__(self, config, compression, path=None):
        """

        :param config: barman.config.ServerConfig
        :param compression: str compression name
        :param path: str|None
        """
        super(PigzCompressor, self).__init__(config, compression, path)
        self._compress = self._build_command("pigz -c -%s" % self.level)
        self._decompress = self._build_command("pigz -c -d")


class PyBZip2Compressor(InternalCompressor):
    """
    Predefined compressor with BZip2 Python libraries
    """

    EXTENSION = ".bz2"
    MAGIC = b"\x42\x5a\x68"
    LEVEL_MIN = 1
    LEVEL_MAX = 9
    LEVEL_LOW = 1
    LEVEL_MEDIUM = 5
    LEVEL_HIGH = 9

    def _compressor(self, name):
        return bz2.BZ2File(name, mode="wb", compresslevel=self.level)

    def _decompressor(self, name):
        return bz2.BZ2File(name, mode="rb")

    def compress_in_mem(self, fileobj):
        in_mem_bz2 = BytesIO(bz2.compress(fileobj.read(), compresslevel=self.level))
        in_mem_bz2.seek(0)
        return in_mem_bz2

    def decompress_in_mem(self, fileobj):
        return bz2.BZ2File(fileobj, "rb")


class XZCompressor(InternalCompressor):
    """
    Predefined compressor with XZ Python library
    """

    EXTENSION = ".xz"
    MAGIC = b"\xfd7zXZ\x00"
    LEVEL_MIN = 1
    LEVEL_MAX = 9
    LEVEL_LOW = 1
    LEVEL_MEDIUM = 3
    LEVEL_HIGH = 5

    def _compressor(self, dst):
        return lzma.open(dst, mode="wb", preset=self.level)

    def _decompressor(self, src):
        return lzma.open(src, mode="rb")

    def compress_in_mem(self, fileobj):
        in_mem_xz = BytesIO(lzma.compress(fileobj.read(), preset=self.level))
        in_mem_xz.seek(0)
        return in_mem_xz

    def decompress_in_mem(self, fileobj):
        return lzma.open(fileobj, "rb")


def _try_import_zstd():
    try:
        import zstandard
    except ImportError:
        raise SystemExit("Missing required python module: zstandard")
    return zstandard


class ZSTDCompressor(InternalCompressor):
    """
    Predefined compressor with zstd
    """

    EXTENSION = ".zst"
    MAGIC = b"(\xb5/\xfd"
    LEVEL_MIN = -22
    LEVEL_MAX = 22
    LEVEL_LOW = 1
    LEVEL_MEDIUM = 4
    LEVEL_HIGH = 9

    def __init__(self, config, compression, path=None):
        """
        Constructor.
        :param config: barman.config.ServerConfig
        :param compression: str compression name
        :param path: str|None
        """
        super(ZSTDCompressor, self).__init__(config, compression, path)
        self._zstd = None

    @property
    def zstd(self):
        if self._zstd is None:
            self._zstd = _try_import_zstd()
        return self._zstd

    def _compressor(self, dst):
        return self.zstd.ZstdCompressor(level=self.level).stream_writer(
            open(dst, mode="wb")
        )

    def _decompressor(self, src):
        return self.zstd.ZstdDecompressor().stream_reader(open(src, mode="rb"))

    def compress_in_mem(self, fileobj):
        in_mem_zstd = BytesIO()
        self.zstd.ZstdCompressor(level=self.level).copy_stream(fileobj, in_mem_zstd)
        in_mem_zstd.seek(0)
        return in_mem_zstd

    def decompress_in_mem(self, fileobj):
        return self.zstd.ZstdDecompressor().stream_reader(fileobj)


def _try_import_lz4():
    try:
        import lz4.frame
    except ImportError:
        raise SystemExit("Missing required python module: lz4")
    return lz4


class LZ4Compressor(InternalCompressor):
    """
    Predefined compressor with lz4
    """

    EXTENSION = ".lz4"
    MAGIC = b"\x04\x22\x4d\x18"
    LEVEL_MIN = 0
    LEVEL_MAX = 16
    LEVEL_LOW = 0
    LEVEL_MEDIUM = 6
    LEVEL_HIGH = 10

    def __init__(self, config, compression, path=None):
        """
        Constructor.
        :param config: barman.config.ServerConfig
        :param compression: str compression name
        :param path: str|None
        """
        super(LZ4Compressor, self).__init__(config, compression, path)
        self._lz4 = None

    @property
    def lz4(self):
        if self._lz4 is None:
            self._lz4 = _try_import_lz4()
        return self._lz4

    def _compressor(self, dst):
        return self.lz4.frame.open(dst, mode="wb", compression_level=self.level)

    def _decompressor(self, src):
        return self.lz4.frame.open(src, mode="rb")

    def compress_in_mem(self, fileobj):
        in_mem_lz4 = BytesIO(
            self.lz4.frame.compress(fileobj.read(), compression_level=self.level)
        )
        in_mem_lz4.seek(0)
        return in_mem_lz4

    def decompress_in_mem(self, fileobj):
        return self.lz4.frame.open(fileobj, mode="rb")


def _try_import_snappy():
    try:
        import snappy
    except ImportError:
        raise SystemExit("Missing required python module: python-snappy")
    return snappy


class SnappyCompressor(InternalCompressor):

    EXTENSION = ".snappy"
    MAGIC = b"\xff\x06\x00\x00sNaPpY"

    def __init__(self, config, compression, path=None):
        """
        Constructor.
        :param config: barman.config.ServerConfig
        :param compression: str compression name
        :param path: str|None
        """
        super(SnappyCompressor, self).__init__(config, compression, path)
        self._snappy = None

    @property
    def snappy(self):
        if self._snappy is None:
            self._snappy = _try_import_snappy()
        return self._snappy

    def _compressor(self, dst):
        """Snappy library does not provide an interface which returns file-objects"""
        return None

    def _decompressor(self, src):
        """Snappy library does not provide an interface which returns file-objects"""
        return None

    def compress(self, src, dst):
        """
        Snappy-compress the source file-object to the destination file-object

        :param src: source file to compress
        :param dst: destination of the decompression
        """
        try:
            with open(src, "rb") as istream:
                with open(dst, "wb") as ostream:
                    compressed_fileobj = self.compress_in_mem(istream)
                    shutil.copyfileobj(compressed_fileobj, ostream)
        except Exception as e:
            raise CommandFailedException(dict(ret=None, err=force_str(e), out=None))
        return 0

    def decompress(self, src, dst):
        """
        Decompress the source file-object to the destination file-object

        :param src: source file to decompress
        :param dst: destination of the decompression
        """
        try:
            with open(src, "rb") as istream:
                with open(dst, "wb") as ostream:
                    decompressed_fileobj = self.decompress_in_mem(istream)
                    shutil.copyfileobj(decompressed_fileobj, ostream)
        except Exception as e:
            raise CommandFailedException(dict(ret=None, err=force_str(e), out=None))
        return 0

    def compress_in_mem(self, fileobj):
        in_mem_snappy = BytesIO()
        self.snappy.stream_compress(fileobj, in_mem_snappy)
        in_mem_snappy.seek(0)
        return in_mem_snappy

    def decompress_in_mem(self, fileobj):
        decompressed_file = BytesIO()
        self.snappy.stream_decompress(fileobj, decompressed_file)
        decompressed_file.seek(0)
        return decompressed_file

    def decompress_to_fileobj(self, src_fileobj, dest_fileobj):
        """
        Decompresses the given file-object on the especified file-object

        :param src_fileobj: source file-object to be decompressed
        :param dest_fileobj: destination file-object to have the decompressed content

        .. note::
            We override this method to avoid redundant work. As Snappy can stream the
            result directly to a specified object, there is no need for intermediate
            objects as used in the parent class implementation.
        """
        self.snappy.stream_decompress(src_fileobj, dest_fileobj)


class CustomCompressor(CommandCompressor):
    """
    Custom compressor
    """

    def __init__(self, config, compression, path=None):
        """

        :param config: barman.config.ServerConfig
        :param compression: str compression name
        :param path: str|None
        """
        if config.custom_compression_filter is None or not isinstance(
            config.custom_compression_filter, str
        ):
            raise CompressionIncompatibility("custom_compression_filter")
        if config.custom_decompression_filter is None or not isinstance(
            config.custom_decompression_filter, str
        ):
            raise CompressionIncompatibility("custom_decompression_filter")

        super(CustomCompressor, self).__init__(config, compression, path)
        self._compress = self._build_command(config.custom_compression_filter)
        self._decompress = self._build_command(config.custom_decompression_filter)


# Maps supported compression methods to their corresponding classes
# NOTE:
# - Entries are sorted alphabetically, so if multiple methods share the same magic number
#   (e.g., 'gzip' and 'pigz'), the first one alphabetically ('gzip') will be chosen.
# - gzip maps to the same class as pygzip, and bzip2 maps to the same class as pybzip2.
#   Previously, they mapped to their own CommandCompressor classes, but invoking
#   subprocesses for compression is something that we want to get away from Barman.
compression_registry = {
    "gzip": PyGZipCompressor,
    "bzip2": PyBZip2Compressor,
    "pygzip": PyGZipCompressor,
    "pybzip2": PyBZip2Compressor,
    "pigz": PigzCompressor,
    "xz": XZCompressor,
    "zstd": ZSTDCompressor,
    "lz4": LZ4Compressor,
    "snappy": SnappyCompressor,
    "custom": CustomCompressor,
}


def get_pg_basebackup_compression(server):
    """
    Factory method which returns an instantiated PgBaseBackupCompression subclass
    for the backup_compression option in config for the supplied server.

    :param barman.server.Server server: the server for which the PgBaseBackupCompression
        should be constructed
    :return GZipPgBaseBackupCompression:
    """
    if server.config.backup_compression is None:
        return
    pg_base_backup_cfg = PgBaseBackupCompressionConfig(
        server.config.backup_compression,
        server.config.backup_compression_format,
        server.config.backup_compression_level,
        server.config.backup_compression_location,
        server.config.backup_compression_workers,
    )
    base_backup_compression_option = None
    compression = None
    if server.config.backup_compression == GZipCompression.name:
        # Create PgBaseBackupCompressionOption
        base_backup_compression_option = GZipPgBaseBackupCompressionOption(
            pg_base_backup_cfg
        )
        compression = GZipCompression(unix_command_factory())

    if server.config.backup_compression == LZ4Compression.name:
        base_backup_compression_option = LZ4PgBaseBackupCompressionOption(
            pg_base_backup_cfg
        )
        compression = LZ4Compression(unix_command_factory())

    if server.config.backup_compression == ZSTDCompression.name:
        base_backup_compression_option = ZSTDPgBaseBackupCompressionOption(
            pg_base_backup_cfg
        )
        compression = ZSTDCompression(unix_command_factory())

    if server.config.backup_compression == NoneCompression.name:
        base_backup_compression_option = NonePgBaseBackupCompressionOption(
            pg_base_backup_cfg
        )
        compression = NoneCompression(unix_command_factory())

    if base_backup_compression_option is None or compression is None:
        # We got to the point where the compression is not handled
        raise CompressionException(
            "Barman does not support pg_basebackup compression: %s"
            % server.config.backup_compression
        )
    return PgBaseBackupCompression(
        pg_base_backup_cfg, base_backup_compression_option, compression
    )


class PgBaseBackupCompressionConfig(object):
    """Should become a dataclass"""

    def __init__(
        self,
        backup_compression,
        backup_compression_format,
        backup_compression_level,
        backup_compression_location,
        backup_compression_workers,
    ):
        self.type = backup_compression
        self.format = backup_compression_format
        self.level = backup_compression_level
        self.location = backup_compression_location
        self.workers = backup_compression_workers


class PgBaseBackupCompressionOption(object):
    """This class is in charge of validating pg_basebackup compression options"""

    def __init__(self, pg_base_backup_config):
        """

        :param pg_base_backup_config: PgBaseBackupCompressionConfig
        """
        self.config = pg_base_backup_config

    def validate(self, pg_server_version, remote_status):
        """
        Validate pg_basebackup compression options.

        :param pg_server_version int: the server for which the
          compression options should be validated.
        :param dict remote_status: the status of the pg_basebackup command
        :return List: List of Issues (str) or empty list
        """
        issues = []
        if self.config.location is not None and self.config.location == "server":
            # "backup_location = server" requires pg_basebackup >= 15
            if remote_status["pg_basebackup_version"] < Version("15"):
                issues.append(
                    "backup_compression_location = server requires "
                    "pg_basebackup 15 or greater"
                )
            # "backup_location = server" requires PostgreSQL >= 15
            if pg_server_version < 150000:
                issues.append(
                    "backup_compression_location = server requires "
                    "PostgreSQL 15 or greater"
                )

        # plain backup format is only allowed when compression is on the server
        if self.config.format == "plain" and self.config.location != "server":
            issues.append(
                "backup_compression_format plain is not compatible with "
                "backup_compression_location %s" % self.config.location
            )
        return issues


class GZipPgBaseBackupCompressionOption(PgBaseBackupCompressionOption):
    def validate(self, pg_server_version, remote_status):
        """
        Validate gzip-specific options.

        :param pg_server_version int: the server for which the
          compression options should be validated.
        :param dict remote_status: the status of the pg_basebackup command
        :return List: List of Issues (str) or empty list
        """
        issues = super(GZipPgBaseBackupCompressionOption, self).validate(
            pg_server_version, remote_status
        )
        levels = list(range(1, 10))
        levels.append(-1)
        if self.config.level is not None and remote_status[
            "pg_basebackup_version"
        ] < Version("15"):
            # version prior to 15 allowed gzip compression 0
            levels.append(0)
            if self.config.level not in levels:
                issues.append(
                    "backup_compression_level %d unsupported by compression algorithm."
                    " %s expects a compression level between -1 and 9 (-1 will use default compression level)."
                    % (self.config.level, self.config.type)
                )
        if (
            self.config.level is not None
            and remote_status["pg_basebackup_version"] >= Version("15")
            and self.config.level not in levels
        ):
            msg = (
                "backup_compression_level %d unsupported by compression algorithm."
                " %s expects a compression level between 1 and 9 (-1 will use default compression level)."
                % (self.config.level, self.config.type)
            )
            if self.config.level == 0:
                msg += "\nIf you need to create an archive not compressed, you should set `backup_compression = none`."
            issues.append(msg)
        if self.config.workers is not None:
            issues.append(
                "backup_compression_workers is not compatible with compression %s"
                % self.config.type
            )
        return issues


class LZ4PgBaseBackupCompressionOption(PgBaseBackupCompressionOption):
    def validate(self, pg_server_version, remote_status):
        """
        Validate lz4-specific options.

        :param pg_server_version int: the server for which the
          compression options should be validated.
        :param dict remote_status: the status of the pg_basebackup command
        :return List: List of Issues (str) or empty list
        """
        issues = super(LZ4PgBaseBackupCompressionOption, self).validate(
            pg_server_version, remote_status
        )
        # "lz4" compression requires pg_basebackup >= 15
        if remote_status["pg_basebackup_version"] < Version("15"):
            issues.append(
                "backup_compression = %s requires "
                "pg_basebackup 15 or greater" % self.config.type
            )

        if self.config.level is not None and (
            self.config.level < 0 or self.config.level > 12
        ):
            issues.append(
                "backup_compression_level %d unsupported by compression algorithm."
                " %s expects a compression level between 1 and 12 (0 will use default compression level)."
                % (self.config.level, self.config.type)
            )
        if self.config.workers is not None:
            issues.append(
                "backup_compression_workers is not compatible with compression %s."
                % self.config.type
            )
        return issues


class ZSTDPgBaseBackupCompressionOption(PgBaseBackupCompressionOption):
    def validate(self, pg_server_version, remote_status):
        """
        Validate zstd-specific options.

        :param pg_server_version int: the server for which the
          compression options should be validated.
        :param dict remote_status: the status of the pg_basebackup command
        :return List: List of Issues (str) or empty list
        """
        issues = super(ZSTDPgBaseBackupCompressionOption, self).validate(
            pg_server_version, remote_status
        )
        # "zstd" compression requires pg_basebackup >= 15
        if remote_status["pg_basebackup_version"] < Version("15"):
            issues.append(
                "backup_compression = %s requires "
                "pg_basebackup 15 or greater" % self.config.type
            )

        # Minimal config level comes from zstd library `STD_minCLevel()` and is
        # commonly set to -131072.
        if self.config.level is not None and (
            self.config.level < -131072 or self.config.level > 22
        ):
            issues.append(
                "backup_compression_level %d unsupported by compression algorithm."
                " '%s' expects a compression level between -131072 and 22 (3 will use default compression level)."
                % (self.config.level, self.config.type)
            )
        if self.config.workers is not None and (
            type(self.config.workers) is not int or self.config.workers < 0
        ):
            issues.append(
                "backup_compression_workers should be a positive integer: '%s' is invalid."
                % self.config.workers
            )
        return issues


class NonePgBaseBackupCompressionOption(PgBaseBackupCompressionOption):
    def validate(self, pg_server_version, remote_status):
        """
        Validate none compression specific options.

        :param pg_server_version int: the server for which the
          compression options should be validated.
        :param dict remote_status: the status of the pg_basebackup command
        :return List: List of Issues (str) or empty list
        """
        issues = super(NonePgBaseBackupCompressionOption, self).validate(
            pg_server_version, remote_status
        )

        if self.config.level is not None and (self.config.level != 0):
            issues.append(
                "backup_compression %s only supports backup_compression_level 0."
                % self.config.type
            )
        if self.config.workers is not None:
            issues.append(
                "backup_compression_workers is not compatible with compression '%s'."
                % self.config.type
            )
        return issues


class PgBaseBackupCompression(object):
    """
    Represents the pg_basebackup compression options and provides functionality
    required by the backup process which depends on those options.
    This is a facade that interacts with appropriate classes
    """

    def __init__(
        self,
        pg_basebackup_compression_cfg,
        pg_basebackup_compression_option,
        compression,
    ):
        """
        Constructor for the PgBaseBackupCompression facade that handles base_backup class related.

        :param pg_basebackup_compression_cfg PgBaseBackupCompressionConfig: pg_basebackup compression  configuration
        :param pg_basebackup_compression_option PgBaseBackupCompressionOption:
        :param compression Compression:

        """
        self.config = pg_basebackup_compression_cfg
        self.options = pg_basebackup_compression_option
        self.compression = compression

    def with_suffix(self, basename):
        """
        Append the suffix to the supplied basename.

        :param str basename: The basename (without compression suffix) of the
          file to be opened.
        """
        return "%s.%s" % (basename, self.compression.file_extension)

    def get_file_content(self, filename, archive):
        """
        Returns archive specific file content
        :param filename: str
        :param archive: str
        :return: str
        """
        return self.compression.get_file_content(filename, archive)

    def validate(self, pg_server_version, remote_status):
        """
        Validate pg_basebackup compression options.

        :param pg_server_version int: the server for which the
          compression options should be validated.
        :param dict remote_status: the status of the pg_basebackup command
        :return List: List of Issues (str) or empty list
        """
        return self.options.validate(pg_server_version, remote_status)


class Compression(with_metaclass(ABCMeta, object)):
    """
    Abstract class meant to represent compression interface
    """

    @abstractproperty
    def name(self):
        """

        :return:
        """

    @abstractproperty
    def file_extension(self):
        """

        :return:
        """

    @abstractmethod
    def decompress(self, src, dst, exclude=None, include_args=None):
        """

        :param src: source file path without compression extension
        :param dst: destination path
        :param exclude: list of filepath in the archive to exclude from the extraction
        :param include_args: list of filepath in the archive to extract.
        :return:
        """

    @abstractmethod
    def get_file_content(self, filename, archive):
        """

        :param filename: str file to search for in the archive (requires its full path within the archive)
        :param archive: str archive path/name without extension
        :return: string content
        """

    def validate_src_and_dst(self, src):
        if src is None or src == "":
            raise ValueError("Source path should be a string")

    def validate_dst(self, dst):
        if dst is None or dst == "":
            raise ValueError("Destination path should be a string")


class GZipCompression(Compression):
    name = "gzip"
    file_extension = "tar.gz"

    def __init__(self, command):
        """

        :param command: barman.fs.UnixLocalCommand
        """
        self.command = command

    def decompress(self, src, dst, exclude=None, include_args=None):
        """

        :param src: source file path without compression extension
        :param dst: destination path
        :param exclude: list of filepath in the archive to exclude from the extraction
        :param include_args: list of filepath in the archive to extract.
        :return:
        """
        self.validate_dst(src)
        self.validate_dst(dst)
        exclude = [] if exclude is None else exclude
        exclude_args = []
        for name in exclude:
            exclude_args.append("--exclude")
            exclude_args.append(name)
        include_args = [] if include_args is None else include_args
        args = ["-xzf", src, "--directory", dst]
        args.extend(exclude_args)
        args.extend(include_args)
        ret = self.command.cmd("tar", args=args)
        out, err = self.command.get_last_output()
        if ret != 0:
            raise CommandFailedException(
                "Error decompressing %s into %s: %s" % (src, dst, err)
            )
        else:
            return self.command.get_last_output()

    def get_file_content(self, filename, archive):
        """

        :param filename: str file to search for in the archive (requires its full path within the archive)
        :param archive: str archive path/name without extension
        :return: string content
        """
        full_archive_name = "%s.%s" % (archive, self.file_extension)
        args = ["-xzf", full_archive_name, "-O", filename, "--occurrence"]
        ret = self.command.cmd("tar", args=args)
        out, err = self.command.get_last_output()
        if ret != 0:
            if "Not found in archive" in err:
                raise FileNotFoundException(
                    err + "archive name: %s" % full_archive_name
                )
            else:
                raise CommandFailedException(
                    "Error reading %s into archive %s: (%s)"
                    % (filename, full_archive_name, err)
                )
        else:
            return out


class LZ4Compression(Compression):
    name = "lz4"
    file_extension = "tar.lz4"

    def __init__(self, command):
        """

        :param command: barman.fs.UnixLocalCommand
        """
        self.command = command

    def decompress(self, src, dst, exclude=None, include_args=None):
        """

        :param src: source file path without compression extension
        :param dst: destination path
        :param exclude: list of filepath in the archive to exclude from the extraction
        :param include_args: list of filepath in the archive to extract.
        :return:
        """
        self.validate_dst(src)
        self.validate_dst(dst)
        exclude = [] if exclude is None else exclude
        exclude_args = []
        for name in exclude:
            exclude_args.append("--exclude")
            exclude_args.append(name)
        include_args = [] if include_args is None else include_args
        args = ["--use-compress-program", "lz4", "-xf", src, "--directory", dst]
        args.extend(exclude_args)
        args.extend(include_args)
        ret = self.command.cmd("tar", args=args)
        out, err = self.command.get_last_output()
        if ret != 0:
            raise CommandFailedException(
                "Error decompressing %s into %s: %s" % (src, dst, err)
            )
        else:
            return self.command.get_last_output()

    def get_file_content(self, filename, archive):
        """

        :param filename: str file to search for in the archive (requires its full path within the archive)
        :param archive: str archive path/name without extension
        :return: string content
        """
        full_archive_name = "%s.%s" % (archive, self.file_extension)
        args = [
            "--use-compress-program",
            "lz4",
            "-xf",
            full_archive_name,
            "-O",
            filename,
            "--occurrence",
        ]
        ret = self.command.cmd("tar", args=args)
        out, err = self.command.get_last_output()
        if ret != 0:
            if "Not found in archive" in err:
                raise FileNotFoundException(
                    err + "archive name: %s" % full_archive_name
                )
            else:
                raise CommandFailedException(
                    "Error reading %s into archive %s: (%s)"
                    % (filename, full_archive_name, err)
                )
        else:
            return out


class ZSTDCompression(Compression):
    name = "zstd"
    file_extension = "tar.zst"

    def __init__(self, command):
        """

        :param command: barman.fs.UnixLocalCommand
        """
        self.command = command

    def decompress(self, src, dst, exclude=None, include_args=None):
        """

        :param src: source file path without compression extension
        :param dst: destination path
        :param exclude: list of filepath in the archive to exclude from the extraction
        :param include_args: list of filepath in the archive to extract.
        :return:
        """
        self.validate_dst(src)
        self.validate_dst(dst)
        exclude = [] if exclude is None else exclude
        exclude_args = []
        for name in exclude:
            exclude_args.append("--exclude")
            exclude_args.append(name)
        include_args = [] if include_args is None else include_args
        args = ["--use-compress-program", "zstd", "-xf", src, "--directory", dst]
        args.extend(exclude_args)
        args.extend(include_args)
        ret = self.command.cmd("tar", args=args)
        out, err = self.command.get_last_output()
        if ret != 0:
            raise CommandFailedException(
                "Error decompressing %s into %s: %s" % (src, dst, err)
            )
        else:
            return self.command.get_last_output()

    def get_file_content(self, filename, archive):
        """

        :param filename: str file to search for in the archive (requires its full path within the archive)
        :param archive: str archive path/name without extension
        :return: string content
        """
        full_archive_name = "%s.%s" % (archive, self.file_extension)
        args = [
            "--use-compress-program",
            "zstd",
            "-xf",
            full_archive_name,
            "-O",
            filename,
            "--occurrence",
        ]
        ret = self.command.cmd("tar", args=args)
        out, err = self.command.get_last_output()
        if ret != 0:
            if "Not found in archive" in err:
                raise FileNotFoundException(
                    err + "archive name: %s" % full_archive_name
                )
            else:
                raise CommandFailedException(
                    "Error reading %s into archive %s: (%s)"
                    % (filename, full_archive_name, err)
                )
        else:
            return out


class NoneCompression(Compression):
    name = "none"
    file_extension = "tar"

    def __init__(self, command):
        """

        :param command: barman.fs.UnixLocalCommand
        """
        self.command = command

    def decompress(self, src, dst, exclude=None, include_args=None):
        """

        :param src: source file path without compression extension
        :param dst: destination path
        :param exclude: list of filepath in the archive to exclude from the extraction
        :param include_args: list of filepath in the archive to extract.
        :return:
        """
        self.validate_dst(src)
        self.validate_dst(dst)
        exclude = [] if exclude is None else exclude
        exclude_args = []
        for name in exclude:
            exclude_args.append("--exclude")
            exclude_args.append(name)
        include_args = [] if include_args is None else include_args
        args = ["-xf", src, "--directory", dst]
        args.extend(exclude_args)
        args.extend(include_args)
        ret = self.command.cmd("tar", args=args)
        out, err = self.command.get_last_output()
        if ret != 0:
            raise CommandFailedException(
                "Error decompressing %s into %s: %s" % (src, dst, err)
            )
        else:
            return self.command.get_last_output()

    def get_file_content(self, filename, archive):
        """

        :param filename: str file to search for in the archive (requires its full path within the archive)
        :param archive: str archive path/name without extension
        :return: string content
        """
        full_archive_name = "%s.%s" % (archive, self.file_extension)
        args = ["-xf", full_archive_name, "-O", filename, "--occurrence"]
        ret = self.command.cmd("tar", args=args)
        out, err = self.command.get_last_output()
        if ret != 0:
            if "Not found in archive" in err:
                raise FileNotFoundException(
                    err + "archive name: %s" % full_archive_name
                )
            else:
                raise CommandFailedException(
                    "Error reading %s into archive %s: (%s)"
                    % (filename, full_archive_name, err)
                )
        else:
            return out


def get_server_config_minimal(compression, compression_level):
    """
    Returns a placeholder for a :class:`~barman.config.ServerConfig` object with all compression
    parameters relevant to :class:`barman.compression.CompressionManager` filled.

    :param str compression: a valid compression algorithm option
    :param str|int|None: a compression level for the specified algorithm
    :return: a fake server config object
    :rtype: SimpleNamespace
    """
    return SimpleNamespace(
        compression=compression,
        compression_level=compression_level,
        custom_compression_magic=None,
        custom_compression_filter=None,
        custom_decompression_filter=None,
    )


def get_internal_compressor(compression, compression_level=None):
    """
    Get a :class:`barman.compression.InternalCompressor`
    for the specified *compression* algorithm

    :param str compression: a valid compression algorithm
    :param str|int|None: a compression level for the specified algorithm
    :return: the respective internal compressor
    :rtype: barman.compression.InternalCompressor
    :raises ValueError: if the compression received is unkown to Barman
    """
    # Replace gzip and bzip2 with their respective internal-compressor options so that
    # we are able to compress/decompress in-memory, avoiding forking an OS process
    if compression == "gzip":
        compression = "pygzip"
    elif compression == "bzip2":
        compression = "pybzip2"
    # Use a fake server config so we can reuse the logic of barman.compression module
    server_config = get_server_config_minimal(compression, compression_level)
    comp_manager = CompressionManager(server_config, None)
    compressor = comp_manager.get_compressor(compression)
    if compressor is None:
        raise ValueError("Unknown compression type: %s" % compression)
    return compressor
